from schematics.types import StringType, DictType, ListType, BooleanType, BaseType


class DataTypeFormatEnum(StringType):
    VALID_TYPE_FORMATS = (
        "int32", "int64",
        "float", "double",
        "byte", "binary", "date", "date-time", "password",
        # additional formats
        "duration", "uuid",
        "file", "uri",
        "arm-id",
    )

    def __init__(self, **kwargs):
        super().__init__(choices=self.VALID_TYPE_FORMATS, **kwargs)


class MimeField(StringType):
    pass


class RegularExpressionField(StringType):
    # This string SHOULD be a valid regular expression
    pass


class SecurityRequirementField(DictType):
    """
    Lists the required security schemes to execute this operation. The object can have multiple security schemes declared in it which are all required (that is, there is a logical AND between the schemes).

    The type of value could be string or list
    """

    def __init__(self, **kwargs):
        super().__init__(
            field=ListType(StringType()), **kwargs)


class ScopesField(DictType):
    """Lists the available scopes for an OAuth2 security scheme."""

    def __init__(self, **kwargs):
        super().__init__(field=StringType(), **kwargs)


class XmsCodeGenerationSettingsField(BaseType):
    """
    x-ms-code-generation-settings extension on info element enables passing code generation settings via the OpenAPI definition.

    https://github.com/Azure/autorest/tree/main/docs/extensions#x-ms-code-generation-settings
    """
    def __init__(self, **kwargs):
        super().__init__(
            default=False,
            serialized_name="x-ms-code-generation-settings",
            deserialize_from="x-ms-code-generation-settings",
            **kwargs
        )


class XcadlGeneratedField(BaseType):

    def __init__(self, **kwargs):
        super().__init__(
            default=False,
            serialized_name="x-cadl-generated",
            deserialize_from="x-cadl-generated",
            **kwargs
        )


class XTypespecGeneratedField(BaseType):

    def __init__(self, **kwargs):
        super().__init__(
            default=False,
            serialized_name="x-typespec-generated",
            deserialize_from="x-typespec-generated",
            **kwargs
        )


class XmsSkipURLEncodingField(BooleanType):
    """
    skips URL encoding for path and query parameters.

    https://github.com/Azure/autorest/tree/main/docs/extensions#x-ms-skip-url-encoding
    """

    def __init__(self, **kwargs):
        super().__init__(
            default=False,
            serialized_name="x-ms-skip-url-encoding",
            deserialize_from="x-ms-skip-url-encoding",
            **kwargs
        )


class XmsParameterLocationField(StringType):
    """
    By default Autorest processes global parameters as properties on the client. For example subscriptionId and apiVersion which are defined in the global parameters section end up being properties of the client. It would be natural to define resourceGroupName once in the global parameters section and then reference it everywhere, rather than repeating the same definition inline everywhere. One may not want resourceGroupName as a property on the client, just because it is defined in the global parameters section. This extension helps you achieve that. You can add this extension with value "method" "x-ms-parameter-location": "method" and resourceGroupName will not be a client property.

    This extension can only be applied on global parameters. If this is applied on any parameter in an operation then it will be ignored.

    https://github.com/Azure/autorest/tree/main/docs/extensions#x-ms-parameter-location
    """

    def __init__(self, **kwargs):
        super().__init__(
            choices=("client", "method"),
            serialized_name="x-ms-parameter-location",
            deserialize_from="x-ms-parameter-location",
            **kwargs
        )


class XmsClientNameField(StringType):
    """
    In some situations, data passed by name, such as query parameters, entity headers, or elements of a JSON document body, are not suitable for use in client-side code. For example, a header like 'x-ms-version' would turn out like xMsVersion, or x_ms_version, or XMsVersion, depending on the preferences of a particular code generator. It may be better to allow a code generator to use 'version' as the name of the parameter in client code.

    By using the 'x-ms-client-name' extension, a name can be defined for use specifically in code generation, separately from the name on the wire. It can be used for query parameters and header parameters, as well as properties of schemas.

    https://github.com/Azure/autorest/tree/main/docs/extensions#x-ms-client-name
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name="x-ms-client-name",
            deserialize_from="x-ms-client-name",
            **kwargs
        )


class XmsExternalField(BooleanType):
    """
    To allow generated clients to share models via shared libraries an x-ms-external extension was introduced. When a Schema Object contains this extensions it's definition will be excluded from generated library. Note that in strongly typed languages the code will not compile unless the assembly containing the type is referenced with the project/library.

    https://github.com/Azure/autorest/tree/main/docs/extensions#x-ms-external
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name="x-ms-external",
            deserialize_from="x-ms-external",
            **kwargs
        )


class XmsDiscriminatorValueField(StringType):
    """
    Swagger 2.0 specification requires that when used, the value of discriminator field MUST match the name of the schema or any schema that inherits it. To overcome this limitation x-ms-discriminator-value extension was introduced.

    https://github.com/Azure/autorest/tree/main/docs/extensions#x-ms-discriminator-value
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name="x-ms-discriminator-value",
            deserialize_from="x-ms-discriminator-value",
            **kwargs
        )


class XmsClientFlattenField(BooleanType):
    """
    This extension allows to flatten deeply nested payloads into a more user friendly object.

    https://github.com/Azure/autorest/tree/main/docs/extensions#x-ms-client-flatten
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name="x-ms-client-flatten",
            deserialize_from="x-ms-client-flatten",
            **kwargs
        )


class MutabilityEnum:

    Create = "create"
    Read = "read"
    Update = "update"


class XmsMutabilityField(ListType):
    """
    This extension offers insight to Autorest on how to generate code (mutability of the property of the model classes being generated). It doesn't alter the modeling of the actual payload that is sent on the wire.

    It is an array of strings with three possible values. The array cannot have repeatable values. Valid values are: "create", "read", "update".
    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-ms-mutability
    """

    VALID_VALUES = (
        MutabilityEnum.Create,  # Indicates that the value of the property can be set while creating/initializing/constructing the object
        MutabilityEnum.Read,   # Indicates that the value of the property can be read
        MutabilityEnum.Update,  # Indicates that value of the property can be updated anytime(even after the object is created)
    )

    def __init__(self, **kwargs):
        super().__init__(
            field=StringType(choices=self.VALID_VALUES),
            serialized_name='x-ms-mutability',
            deserialize_from='x-ms-mutability',
            **kwargs
        )


class XmsExamplesField(DictType):
    """
    Describes the format for specifying examples for request and response of an operation in an OpenAPI definition. It is a dictionary of different variations of the examples for a given operation.

    https://github.com/Azure/azure-rest-api-specs/blob/master/documentation/x-ms-examples.md
    """

    def __init__(self, **kwargs):
        super().__init__(
            field=BaseType(),
            serialized_name='x-ms-examples',
            deserialize_from='x-ms-examples',
            **kwargs
        )


class XmsErrorResponseField(BooleanType):
    """
    Indicates whether the response status code should be treated as an error response or not.

    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-ms-error-response
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-ms-error-response',
            deserialize_from='x-ms-error-response',
            **kwargs
        )


class XmsTextField(BooleanType):
    """
    Swagger spec doesn't allow dev to model this XML structure: <title language="text">the title</title> This is well known issue: https://github.com/OAI/OpenAPI-Specification/issues/630

    This extension is defined to help for this scenario.

    Note: The extension is not tight to this particular scenario (you could model any text node that way), but we recommend to follow as much as possible the Swagger specification, and to use it only in case of attributes/text.

    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-ms-text
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-ms-text',
            deserialize_from='x-ms-text',
            **kwargs
        )


class XmsClientDefaultField(BaseType):
    """
    Set the default value for a property or a parameter.

    With this extension, you can set a default value for a property or parameter that is independent of how the property / parameter's schema is handling a default. This is different than the default value you can specify

    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-ms-client-default

    Schema: string, integer, long, float, double, boolean
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-ms-client-default',
            deserialize_from='x-ms-client-default',
            **kwargs
        )


class XmsArmIdDetailsField(BaseType):
    """

    http://azure.github.io/autorest/extensions/#x-ms-arm-id-details
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-ms-arm-id-details',
            deserialize_from='x-ms-arm-id-details',
            **kwargs
        )


class XmsAzureResourceField(BooleanType):
    """
    Resource types as defined by the Resource Manager API are tagged by using a x-ms-azure-resource extension.

    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-ms-azure-resource
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-ms-azure-resource',
            deserialize_from='x-ms-azure-resource',
            **kwargs
        )


class XmsRequestIdField(StringType):
    """
    When set, allows to overwrite the x-ms-request-id response header (default is x-ms-request-id).

    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-ms-request-id
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-ms-request-id',
            deserialize_from='x-ms-request-id',
            **kwargs
        )


class XmsClientRequestIdField(BooleanType):
    """
    When set, specifies the header parameter to be used instead of x-ms-client-request-id (default is x-ms-client-request-id).

    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-ms-client-request-id
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-ms-client-request-id',
            deserialize_from='x-ms-client-request-id',
            **kwargs
        )


class XmsApiVersionField(BooleanType):
    """"""

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-ms-api-version',
            deserialize_from='x-ms-api-version',
            **kwargs
        )


class XmsSkipUrlEncodingField(BooleanType):
    """"""

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-ms-skip-url-encoding',
            deserialize_from='x-ms-skip-url-encoding',
            **kwargs
        )


class XmsSecretField(BooleanType):

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-ms-secret',
            deserialize_from='x-ms-secret',
            **kwargs
        )


class XNullableField(BooleanType):
    """
    Set "x-nullable": true on a schema to indicate that a null is a legal value. By default, a null value should be disallowed when forming a request and rejected during payload deserialization.

    For arrays, sending/receiving a null array entry is not supported and should result in an error.

    https://github.com/Azure/autorest/blob/main/docs/extensions/readme.md#x-nullable
    """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-nullable',
            deserialize_from='x-nullable',
            **kwargs
        )


class XCadlNameField(StringType):
    """ https://github.com/microsoft/cadl """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-cadl-name',
            deserialize_from='x-cadl-name',
            **kwargs
        )


class XTypespecNameField(StringType):

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-typespec-name',
            deserialize_from='x-typespec-name',
            **kwargs
        )


class XmsHeaderCollectionPrefix(StringType):
    """ only used in Storage Data plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-ms-header-collection-prefix',
            deserialize_from='x-ms-header-collection-prefix',
            **kwargs
        )


class XAccessibilityField(StringType):
    """ only used in ContainerRegistry Data plane """

    def __init__(self, **kwargs):
        super().__init__(
            choices=("internal", "public"),
            serialized_name='x-accessibility',
            deserialize_from='x-accessibility',
            **kwargs
        )


class XRequiredField(BooleanType):
    """ only used in ContainerRegistry Data plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-required',
            deserialize_from='x-required',
            **kwargs
        )


class XPublishField(BooleanType):
    """ only used in Maps Data Plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-publish',
            deserialize_from='x-publish',
            **kwargs
        )


class XAzSearchDeprecatedField(BooleanType):
    """ only used in Search Data Plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-az-search-deprecated',
            deserialize_from='x-az-search-deprecated',
            **kwargs
        )


class XSfCodeGenField(BaseType):
    """ only used in ServiceFabricMesh Mgmt Plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-sf-codegen',
            deserialize_from='x-sf-codegen',
            **kwargs
        )


class XSfClientLibField(BaseType):
    """ only used in ServiceFabric Data Plane and ServiceFabricManagedClusters Mgmt Plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-sf-clientlib',
            deserialize_from='x-sf-clientlib',
            **kwargs
        )


class XApimCodeNillableField(BooleanType):
    """ only used in ApiManagement Mgmt Plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-apim-code-nillable',
            deserialize_from='x-apim-code-nillable',
            **kwargs
        )


class XCommentField(StringType):
    """ Only used in IoTCenter Mgmt Plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-comment',
            deserialize_from='x-comment',
            **kwargs
        )


class XOriginalNameField(StringType):
    """ Only used in Marketplane Catalog Data Plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-originalName',
            deserialize_from='x-originalName',
            **kwargs
        )


class XAbstractField(BooleanType):
    """ Only used in Logic Mgmt Plane and Web Mgmt Plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-abstract',
            deserialize_from='x-abstract',
            **kwargs
        )


class XClientNameField(StringType):
    """ Only used in Maps Data Plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-client-name',
            deserialize_from='x-client-name',
            **kwargs
        )


class XNewPatternField(StringType):
    """ Only used in FrontDoor Mgmt Plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-new-pattern',
            deserialize_from='x-new-pattern',
            **kwargs
        )


class XPreviousPatternField(StringType):
    """ Only used in FrontDoor Mgmt Plane """

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-previous-pattern',
            deserialize_from='x-previous-pattern',
            **kwargs
        )


class XADLNameField(StringType):
    """ Only used in Plane"""

    def __init__(self, **kwargs):
        super().__init__(
            serialized_name='x-adl-name',
            deserialize_from='x-adl-name',
            **kwargs
        )



from command.model.configuration import CMDHttpOperation, CMDHttpRequestJsonBody, CMDArraySchema, \
    CMDInstanceUpdateOperation, CMDRequestJson, CMDHttpResponseJsonBody, CMDObjectSchema, CMDSchema, \
    CMDStringSchemaBase, CMDIntegerSchemaBase, CMDFloatSchemaBase, CMDBooleanSchemaBase, CMDObjectSchemaBase, \
    CMDArraySchemaBase, CMDClsSchemaBase, CMDJsonInstanceUpdateAction, CMDObjectSchemaDiscriminator, CMDSchemaEnum, \
    CMDJsonInstanceCreateAction, CMDJsonInstanceDeleteAction, CMDInstanceCreateOperation, CMDInstanceDeleteOperation
from utils import exceptions
from utils.case import to_snack_case
from utils.error_format import AAZErrorFormatEnum


class AzOperationGenerator:

    def __init__(self, name, cmd_ctx, operation):
        self.name = name
        self._cmd_ctx = cmd_ctx
        self._operation = operation
        self.is_long_running = False

    @property
    def when(self):
        return self._operation.when


class AzLifeCycleCallbackGenerator(AzOperationGenerator):

    def __init__(self, name):
        super().__init__(name, None, None)

    @property
    def when(self):
        return None


class AzLifeCycleInstanceUpdateCallbackGenerator(AzLifeCycleCallbackGenerator):

    def __init__(self, name, variant_key, is_selector_variant):
        super().__init__(name)
        self.variant_key = variant_key
        self.is_selector_variant = is_selector_variant


class AzHttpOperationGenerator(AzOperationGenerator):

    def __init__(self, name, cmd_ctx, operation):
        super().__init__(name, cmd_ctx, operation)
        assert isinstance(self._operation, CMDHttpOperation)

        if self._operation.long_running is not None:
            self.is_long_running = True
            self.lro_options = {
                'final-state-via': self._operation.long_running.final_state_via
            }

        self.success_responses = []
        self.success_202_response = None

        error_format = None
        for response in self._operation.http.responses:
            if not response.is_error:
                if self.is_long_running and response.status_codes == [202]:
                    # ignore 202 response for long running operation.
                    # Long running operation should use 200 or 201 schema to deserialize the final response.
                    continue
                self.success_responses.append(AzHttpResponseGenerator(self._cmd_ctx, response))
            else:
                if not isinstance(response.body, CMDHttpResponseJsonBody):
                    if not response.body:
                        raise exceptions.InvalidAPIUsage(
                            f"Invalid `Error` response schema in operation `{self._operation.operation_id}`: "
                            f"Missing `schema` property in response "
                            f"`{response.status_codes or 'default'}`."
                        )
                    else:
                        raise exceptions.InvalidAPIUsage(
                            f"Invalid `Error` response schema in operation `{self._operation.operation_id}`: "
                            f"Only support json schema, current is '{type(response.body)}' in response "
                            f"`{response.status_codes or 'default'}`"
                        )
                schema = response.body.json.schema
                if not isinstance(schema, CMDClsSchemaBase):
                    raise NotImplementedError()
                name = schema.type[1:]
                if not error_format:
                    error_format = name
                if error_format != name:
                    raise exceptions.InvalidAPIUsage(
                        f"Invalid `Error` response schema in operation `{self._operation.operation_id}`: "
                        f"Multiple schema formats are founded: {name}, {error_format}"
                    )
        if not error_format:
            raise exceptions.InvalidAPIUsage(
                f"Missing `Error` response schema in operation `{self._operation.operation_id}`: "
                f"Please define the `default` response in swagger for error."
            )
        elif not AAZErrorFormatEnum.validate(error_format):
            raise exceptions.InvalidAPIUsage(
                f"Invalid `Error` response schema in operation `{self._operation.operation_id}`: "
                f"Invalid error format `{error_format}`. Support `ODataV4Format` and `MgmtErrorFormat` only"
            )

        if self.is_long_running:
            callback_name = None
            for response in self.success_responses:
                if 200 in response.status_codes or 201 in response.status_codes:
                    callback_name = response.callback_name
                    break
            self.success_202_response = AzHttp202ResponseGenerator(self._cmd_ctx, callback_name)
        self.error_format = error_format

        # specify content
        self.content = None
        self.form_content = None
        self.stream_content = None
        if self._operation.http.request.body:
            body = self._operation.http.request.body
            if isinstance(body, CMDHttpRequestJsonBody):
                self.content = AzHttpRequestContentGenerator(self._cmd_ctx, body)
            else:
                raise NotImplementedError()

    @property
    def url(self):
        return self._operation.http.path

    @property
    def method(self):
        return self._operation.http.request.method.upper()

    @property
    def url_parameters(self):
        path = self._operation.http.request.path
        if not path:
            return None
        parameters = []
        if path.params:
            for param in path.params:
                kwargs = {}
                if param.skip_url_encoding:
                    kwargs['skip_quote'] = True
                if param.required:
                    kwargs['required'] = param.required
                arg_key, hide = self._cmd_ctx.get_argument(param.arg)
                if not hide:
                    parameters.append([
                        param.name,
                        arg_key,
                        False,
                        kwargs
                    ])
        if path.consts:
            for param in path.consts:
                assert param.const
                kwargs = {}
                if param.skip_url_encoding:
                    kwargs['skip_quote'] = True
                if param.required:
                    kwargs['required'] = param.required
                parameters.append([
                    param.name,
                    param.default.value,
                    True,
                    kwargs
                ])
        return parameters

    @property
    def query_parameters(self):
        query = self._operation.http.request.query
        if not query:
            return None
        parameters = []
        if query.params:
            for param in query.params:
                kwargs = {}
                if param.skip_url_encoding:
                    kwargs['skip_quote'] = True
                if param.required:
                    kwargs['required'] = param.required
                arg_key, hide = self._cmd_ctx.get_argument(param.arg)
                if not hide:
                    parameters.append([
                        param.name,
                        arg_key,
                        False,
                        kwargs
                    ])
        if query.consts:
            for param in query.consts:
                assert param.const
                kwargs = {}
                if param.skip_url_encoding:
                    kwargs['skip_quote'] = True
                if param.required:
                    kwargs['required'] = param.required
                parameters.append([
                    param.name,
                    param.default.value,
                    True,
                    kwargs
                ])
        return parameters

    @property
    def header_parameters(self):
        header = self._operation.http.request.header
        parameters = []
        if header:
            if header.params:
                for param in header.params:
                    kwargs = {}
                    if param.required:
                        kwargs['required'] = param.required
                    arg_key, hide = self._cmd_ctx.get_argument(param.arg)
                    if not hide:
                        parameters.append([
                            param.name,
                            arg_key,
                            False,
                            kwargs
                        ])
            if header.consts:
                for param in header.consts:
                    assert param.const
                    kwargs = {}
                    if param.required:
                        kwargs['required'] = param.required
                    parameters.append([
                        param.name,
                        param.default.value,
                        True,
                        kwargs
                    ])
        if self.content:
            body = self._operation.http.request.body
            if isinstance(body, CMDHttpRequestJsonBody):
                parameters.append([
                    "Content-Type",
                    "application/json",
                    True,
                    {}
                ])
        if self.success_responses:
            for response in self.success_responses:
                if response._response.body is not None and isinstance(response._response.body, CMDHttpResponseJsonBody):
                    parameters.append([
                        "Accept",
                        "application/json",
                        True,
                        {}
                    ])
                    break
        return parameters


class AzInstanceOperationGenerator(AzOperationGenerator):

    def __init__(self, name, cmd_ctx, operation, arg_key, variant_key, is_selector_variant):
        super().__init__(name, cmd_ctx, operation)
        self.arg_key = arg_key
        self.variant_key = variant_key
        self.is_selector_variant = is_selector_variant


class AzJsonCreateOperationGenerator(AzInstanceOperationGenerator):

    HANDLER_NAME = "_create_instance"

    VALUE_NAME = "_instance_value"

    BUILDER_NAME = "_builder"

    def __init__(self, name, cmd_ctx, operation):
        variant_key, is_selector_variant = cmd_ctx.get_variant(operation.instance_create.ref)

        super().__init__(name, cmd_ctx, operation,
                         arg_key="self.ctx.args",
                         variant_key=variant_key,
                         is_selector_variant=is_selector_variant)

        assert isinstance(self._operation, CMDInstanceCreateOperation)
        assert isinstance(self._operation.instance_create, CMDJsonInstanceCreateAction)
        self._json = self._operation.instance_create.json

        assert self._json.ref is None
        assert isinstance(self._json.schema, CMDSchema)
        if self._json.schema.arg:
            self.arg_key, hide = self._cmd_ctx.get_argument(self._json.schema.arg)
            assert not hide

        self.typ, _, self.cls_builder_name = render_schema(
            self._json.schema, self._cmd_ctx.update_clses, name=self._json.schema.name)

        self._update_over_schema(self._json.schema)

    def _update_over_schema(self, s):
        if getattr(s, 'cls', None):
            self._cmd_ctx.set_update_cls(s)

        if isinstance(s, CMDObjectSchemaBase):
            if s.props:
                for prop in s.props:
                    self._update_over_schema(prop)
            if s.additional_props and s.additional_props.item:
                self._update_over_schema(s.additional_props.item)
            if s.discriminators:
                for disc in s.discriminators:
                    self._update_over_schema(disc)
        elif isinstance(s, CMDArraySchemaBase):
            if s.item:
                self._update_over_schema(s.item)
        elif isinstance(s, CMDObjectSchemaDiscriminator):
            if s.props:
                for prop in s.props:
                    self._update_over_schema(prop)
            if s.discriminators:
                for disc in s.discriminators:
                    self._update_over_schema(disc)

    def iter_scopes(self):
        if not self._json.schema or not isinstance(self._json.schema, (CMDObjectSchema, CMDArraySchema)):
            return

        for scopes in _iter_request_scopes_by_schema_base(self._json.schema, self.BUILDER_NAME, None, self.arg_key, self._cmd_ctx):
            yield scopes


class AzJsonDeleteOperationGenerator(AzInstanceOperationGenerator):

    HANDLER_NAME = "_delete_instance"

    def __init__(self, name, cmd_ctx, operation):
        variant_key, is_selector_variant = cmd_ctx.get_variant(operation.instance_delete.ref)

        super().__init__(name, cmd_ctx, operation,
                         arg_key="self.ctx.args",
                         variant_key=variant_key,
                         is_selector_variant=is_selector_variant)
        assert isinstance(self._operation, CMDInstanceDeleteOperation)
        assert isinstance(self._operation.instance_delete, CMDJsonInstanceDeleteAction)
        self._json = self._operation.instance_delete.json

        assert self._json.ref is None
        assert self._json.schema is None


class AzInstanceUpdateOperationGenerator(AzInstanceOperationGenerator):
    pass


class AzJsonUpdateOperationGenerator(AzInstanceUpdateOperationGenerator):

    HANDLER_NAME = "_update_instance"

    VALUE_NAME = "_instance_value"

    BUILDER_NAME = "_builder"

    def __init__(self, name, cmd_ctx, operation):
        variant_key, is_selector_variant = cmd_ctx.get_variant(operation.instance_update.ref)
        super().__init__(name, cmd_ctx, operation,
                         arg_key="self.ctx.args",
                         variant_key=variant_key,
                         is_selector_variant=is_selector_variant)
        assert isinstance(self._operation, CMDInstanceUpdateOperation)
        assert isinstance(self._operation.instance_update, CMDJsonInstanceUpdateAction)
        self._json = self._operation.instance_update.json

        assert self._json.ref is None
        assert isinstance(self._json.schema, CMDSchema)
        if self._json.schema.arg:
            self.arg_key, hide = self._cmd_ctx.get_argument(self._json.schema.arg)
            assert not hide
        self.typ, _, self.cls_builder_name = render_schema(
            self._json.schema, self._cmd_ctx.update_clses, name=self._json.schema.name)

        self._update_over_schema(self._json.schema)

    def _update_over_schema(self, s):
        if getattr(s, 'cls', None):
            self._cmd_ctx.set_update_cls(s)

        if isinstance(s, CMDObjectSchemaBase):
            if s.props:
                for prop in s.props:
                    self._update_over_schema(prop)
            if s.additional_props and s.additional_props.item:
                self._update_over_schema(s.additional_props.item)
            if s.discriminators:
                for disc in s.discriminators:
                    self._update_over_schema(disc)
        elif isinstance(s, CMDArraySchemaBase):
            if s.item:
                self._update_over_schema(s.item)
        elif isinstance(s, CMDObjectSchemaDiscriminator):
            if s.props:
                for prop in s.props:
                    self._update_over_schema(prop)
            if s.discriminators:
                for disc in s.discriminators:
                    self._update_over_schema(disc)

    def iter_scopes(self):
        if not self._json.schema or not isinstance(self._json.schema, (CMDObjectSchema, CMDArraySchema)):
            return

        for scopes in _iter_request_scopes_by_schema_base(self._json.schema, self.BUILDER_NAME, None, self.arg_key, self._cmd_ctx):
            yield scopes


class AzGenericUpdateOperationGenerator(AzInstanceUpdateOperationGenerator):

    def __init__(self, cmd_ctx, variant_key, is_selector_variant):
        super().__init__("InstanceUpdateByGeneric", cmd_ctx, None,
                         arg_key="self.ctx.generic_update_args",
                         variant_key=variant_key,
                         is_selector_variant=is_selector_variant)

    @property
    def when(self):
        return None


class AzHttpRequestContentGenerator:
    VALUE_NAME = "_content_value"
    BUILDER_NAME = "_builder"

    def __init__(self, cmd_ctx, body):
        self._cmd_ctx = cmd_ctx
        assert isinstance(body.json, CMDRequestJson)
        self._json = body.json
        self.ref = None
        if self._json.ref:
            self.ref, is_selector = self._cmd_ctx.get_variant(self._json.ref)
            assert not is_selector
        self.arg_key = "self.ctx.args"
        if self.ref is None:
            assert isinstance(self._json.schema, CMDSchema)
            if self._json.schema.arg:
                self.arg_key, hide = self._cmd_ctx.get_argument(self._json.schema.arg)
                assert not hide
            self.typ, self.typ_kwargs, self.cls_builder_name = render_schema(
                self._json.schema, self._cmd_ctx.update_clses, name=self._json.schema.name)

            self._update_over_schema(self._json.schema)

    def _update_over_schema(self, s):
        if getattr(s, 'cls', None):
            self._cmd_ctx.set_update_cls(s)

        if isinstance(s, CMDObjectSchemaBase):
            if s.props:
                for prop in s.props:
                    self._update_over_schema(prop)
            if s.additional_props and s.additional_props.item:
                self._update_over_schema(s.additional_props.item)
            if s.discriminators:
                for disc in s.discriminators:
                    self._update_over_schema(disc)
        elif isinstance(s, CMDArraySchemaBase):
            if s.item:
                self._update_over_schema(s.item)
        elif isinstance(s, CMDObjectSchemaDiscriminator):
            if s.props:
                for prop in s.props:
                    self._update_over_schema(prop)
            if s.discriminators:
                for disc in s.discriminators:
                    self._update_over_schema(disc)

    def iter_scopes(self):
        if not self._json.schema or not isinstance(self._json.schema, (CMDObjectSchema, CMDArraySchema)):
            return

        for scopes in _iter_request_scopes_by_schema_base(self._json.schema, self.BUILDER_NAME, None, self.arg_key, self._cmd_ctx):
            yield scopes


class AzRequestClsGenerator:
    BUILDER_NAME = "_builder"

    def __init__(self, cmd_ctx, name, schema):
        self._cmd_ctx = cmd_ctx
        self.schema = schema
        self.name = name
        self.builder_name = parse_cls_builder_name(name)

    def iter_scopes(self):
        arg_key = f"@{self.name}"
        for scopes in _iter_request_scopes_by_schema_base(self.schema, self.BUILDER_NAME, None, arg_key, self._cmd_ctx):
            yield scopes


class AzHttpResponseGenerator:

    @staticmethod
    def _generate_callback_name(status_codes):
        return "on_" + "_".join(str(code) for code in status_codes)

    def __init__(self, cmd_ctx, response):
        self._cmd_ctx = cmd_ctx
        self._response = response

        self.status_codes = response.status_codes
        self.callback_name = self._generate_callback_name(response.status_codes)

        self.variant_name = None
        if response.body is not None and isinstance(response.body, CMDHttpResponseJsonBody) and \
                response.body.json is not None and response.body.json.var is not None:
            variant = response.body.json.var
            self.variant_name = self._cmd_ctx.get_variant(variant, name_only=True)
            self.schema = AzHttpResponseSchemaGenerator(
                self._cmd_ctx, response.body.json.schema, response.status_codes
            )


class AzHttp202ResponseGenerator:

    def __init__(self, cmd_ctx, callback_name):
        self._cmd_ctx = cmd_ctx
        self.status_codes = [202]
        self.callback_name = callback_name


class AzHttpResponseSchemaGenerator:

    @staticmethod
    def _generate_schema_name(status_codes):
        return "_schema_on_" + "_".join(str(code) for code in status_codes)

    @staticmethod
    def _generate_schema_builder_name(status_codes):
        return "_build_schema_on_" + "_".join(str(code) for code in status_codes)

    def __init__(self, cmd_ctx, schema, status_codes):
        self._cmd_ctx = cmd_ctx
        self._schema = schema
        self.name = self._generate_schema_name(status_codes)
        self.builder_name = self._generate_schema_builder_name(status_codes)

        self.typ, self.typ_kwargs, self.cls_builder_name = render_schema_base(self._schema, self._cmd_ctx.response_clses)

        self._update_over_schema(self._schema)

    def _update_over_schema(self, s):
        if getattr(s, 'cls', None):
            self._cmd_ctx.set_response_cls(s)

        if isinstance(s, CMDObjectSchemaBase):
            if s.props:
                for prop in s.props:
                    self._update_over_schema(prop)
            if s.additional_props and s.additional_props.item:
                self._update_over_schema(s.additional_props.item)
            if s.discriminators:
                for disc in s.discriminators:
                    self._update_over_schema(disc)
        elif isinstance(s, CMDArraySchemaBase):
            if s.item:
                self._update_over_schema(s.item)
        elif isinstance(s, CMDObjectSchemaDiscriminator):
            if s.props:
                for prop in s.props:
                    self._update_over_schema(prop)
            if s.discriminators:
                for disc in s.discriminators:
                    self._update_over_schema(disc)

    def iter_scopes(self):
        if not self.cls_builder_name and isinstance(self._schema, (CMDObjectSchemaBase, CMDArraySchemaBase)):
            for scopes in _iter_response_scopes_by_schema_base(self._schema, self.name, f"cls.{self.name}", self._cmd_ctx):
                yield scopes


class AzResponseClsGenerator:

    @staticmethod
    def _generate_schema_name(name):
        return "_schema_" + to_snack_case(name)

    def __init__(self, cmd_ctx, name, schema):
        self._cmd_ctx = cmd_ctx
        self.schema = schema
        self.name = name
        self.schema_name = self._generate_schema_name(name)
        self.builder_name = parse_cls_builder_name(name)

        self.typ, self.typ_kwargs, _ = render_schema_base(self.schema, self._cmd_ctx.response_clses)

        self.props = []
        self.discriminators = []
        if isinstance(schema, CMDObjectSchemaBase):
            if schema.props and schema.additional_props:
                raise NotImplementedError()
            if schema.discriminators:
                for disc in schema.discriminators:
                    disc_key = to_snack_case(disc.property)
                    disc_value = disc.value
                    self.discriminators.append((disc_key, disc_value))
            if schema.props:
                for s in schema.props:
                    s_name = to_snack_case(s.name)
                    self.props.append(s_name)
            elif schema.additional_props:
                if schema.additional_props.item is not None:
                    self.props.append("Element")
                else:
                    assert schema.additional_props.any_type is True
            else:
                raise NotImplementedError()
        elif isinstance(schema, CMDArraySchemaBase):
            self.props.append("Element")

        self.props = sorted(self.props)

    def iter_scopes(self):
        for scopes in _iter_response_scopes_by_schema_base(self.schema, to_snack_case(self.name), self.schema_name, self._cmd_ctx):
            yield scopes


def _iter_request_scopes_by_schema_base(schema, name, scope_define, arg_key, cmd_ctx):
    rendered_schemas = []
    search_schemas = {}
    discriminators = []
    if isinstance(schema, CMDObjectSchemaBase):
        if schema.props and schema.additional_props:
            # not support for both props and additional props
            raise NotImplementedError()

        if schema.discriminators:
            discriminators.extend(schema.discriminators)

        if schema.props:
            for s in schema.props:
                s_name = s.name
                s_typ, s_typ_kwargs, cls_builder_name = render_schema(s, cmd_ctx.update_clses, s_name)
                if s.arg:
                    # current schema linked with argument
                    s_arg_key, hide = cmd_ctx.get_argument(s.arg)
                    if hide:
                        continue
                else:
                    # current schema not linked with argument, then use current arg_key
                    s_arg_key = arg_key

                # handle enum item attached with argument
                has_enum_argument = False
                if hasattr(s, "enum") and isinstance(s.enum, CMDSchemaEnum):
                    for item in s.enum.items:
                        if item.arg:
                            # enum item linked with argument
                            item_arg_key, hide = cmd_ctx.get_argument(item.arg)
                            if hide:
                                continue
                            has_enum_argument = True
                            r_key = item_arg_key.replace(arg_key, '')
                            if not r_key:
                                r_key = '.'  # which means if the arg exist, fill it
                            is_const = True
                            const_value = item.value
                            rendered_schemas.append(
                                (s_name, s_typ, is_const, const_value, r_key, s_typ_kwargs, cls_builder_name)
                            )

                if not has_enum_argument:
                    r_key = s_arg_key.replace(arg_key, '')
                    if not r_key:
                        r_key = '.' if s.required else None  # which means if the parent exist, fill it

                    is_const = False
                    const_value = None
                    if s.const:
                        is_const = True
                        const_value = s.default.value

                    rendered_schemas.append(
                        (s_name, s_typ, is_const, const_value, r_key, s_typ_kwargs, cls_builder_name)
                    )
                    if not cls_builder_name and not is_const \
                            and isinstance(s, (CMDObjectSchemaBase, CMDArraySchemaBase)):
                        search_schemas[s_name] = (s, s_arg_key)

        elif schema.additional_props:
            if schema.additional_props.any_type is True:
                s_name = '{}'
                r_key = '.'  # if element exist, always fill it

                is_const = False
                const_value = None
                rendered_schemas.append(
                    (s_name, None, is_const, const_value, r_key, None, None)
                )
            else:
                assert schema.additional_props.item is not None
                s = schema.additional_props.item
                s_name = '{}'
                s_typ, s_typ_kwargs, cls_builder_name = render_schema_base(s, cmd_ctx.update_clses)
                s_arg_key = arg_key + '{}'
                r_key = '.'  # if element exist, always fill it

                is_const = False
                const_value = None
                rendered_schemas.append(
                    (s_name, s_typ, is_const, const_value, r_key, s_typ_kwargs, cls_builder_name)
                )
                if not cls_builder_name and isinstance(s, (CMDObjectSchemaBase, CMDArraySchemaBase)):
                    search_schemas[s_name] = (s, s_arg_key)

    elif isinstance(schema, CMDObjectSchemaDiscriminator):
        if schema.discriminators:
            discriminators.extend(schema.discriminators)

        if schema.props:
            for s in schema.props:
                s_name = s.name
                s_typ, s_typ_kwargs, cls_builder_name = render_schema(s, cmd_ctx.update_clses, s_name)
                if s.arg:
                    # current schema linked with argument
                    s_arg_key, hide = cmd_ctx.get_argument(s.arg)
                    if hide:
                        continue
                else:
                    # current schema not linked with argument, then use current arg_key
                    s_arg_key = arg_key

                # handle enum item attached with argument
                has_enum_argument = False
                if hasattr(s, "enum") and isinstance(s.enum, CMDSchemaEnum):
                    for item in s.enum.items:
                        if item.arg:
                            # enum item linked with argument
                            item_arg_key, hide = cmd_ctx.get_argument(item.arg)
                            if hide:
                                continue
                            has_enum_argument = True
                            r_key = item_arg_key.replace(arg_key, '')
                            if not r_key:
                                r_key = '.'
                            is_const = True
                            const_value = item.value
                            rendered_schemas.append(
                                (s_name, s_typ, is_const, const_value, r_key, s_typ_kwargs, cls_builder_name)
                            )

                if not has_enum_argument:
                    r_key = s_arg_key.replace(arg_key, '')
                    if not r_key:
                        r_key = '.' if s.required else None

                    is_const = False
                    const_value = None
                    if s.const:
                        is_const = True
                        const_value = s.default.value

                    rendered_schemas.append(
                        (s_name, s_typ, is_const, const_value, r_key, s_typ_kwargs, cls_builder_name)
                    )
                    if not cls_builder_name and not is_const \
                            and isinstance(s, (CMDObjectSchemaBase, CMDArraySchemaBase)):
                        # step into object and array schemas
                        search_schemas[s_name] = (s, s_arg_key)

    elif isinstance(schema, CMDArraySchemaBase):
        assert schema.item is not None  # make sure array schema defined its element schema
        s = schema.item
        s_name = '[]'
        s_typ, s_typ_kwargs, cls_builder_name = render_schema_base(s, cmd_ctx.update_clses)
        s_arg_key = arg_key + '[]'
        r_key = '.'  # if element exist, always fill it

        is_const = False
        const_value = None
        rendered_schemas.append(
            (s_name, s_typ, is_const, const_value, r_key, s_typ_kwargs, cls_builder_name)
        )
        if not cls_builder_name and isinstance(s, (CMDObjectSchemaBase, CMDArraySchemaBase)):
            search_schemas[s_name] = (s, s_arg_key)

    else:
        raise NotImplementedError()

    if rendered_schemas or discriminators:
        disc_defines = [(disc.property, disc.value) for disc in discriminators]
        yield name, scope_define, rendered_schemas, disc_defines

    scope_define = scope_define or ""
    for s_name, (s, s_arg_key) in search_schemas.items():
        if s_name == '[]':
            s_scope_define = scope_define + '[]'
            s_name = '_elements'
        elif s_name == '{}':
            s_scope_define = scope_define + '{}'
            s_name = '_elements'
        else:
            s_scope_define = f"{scope_define}.{s_name}"
        for scopes in _iter_request_scopes_by_schema_base(s, to_snack_case(s_name), s_scope_define, s_arg_key, cmd_ctx):
            yield scopes

    for disc in discriminators:
        key_name = disc.property
        key_value = disc.value
        disc_name = f"disc_{to_snack_case(disc.value)}"
        disc_scope_define = scope_define + "{" + key_name + ":" + key_value + "}"
        disc_arg_key = arg_key
        for scopes in _iter_request_scopes_by_schema_base(disc, disc_name, disc_scope_define, disc_arg_key, cmd_ctx):
            yield scopes


def _iter_response_scopes_by_schema_base(schema, name, scope_define, cmd_ctx):
    rendered_schemas = []
    search_schemas = {}
    discriminators = []
    if isinstance(schema, CMDObjectSchemaBase):
        if schema.props and schema.additional_props:
            # not support to parse schema with both props and additional_props
            raise NotImplementedError()

        if schema.discriminators:
            discriminators.extend(schema.discriminators)

        if schema.props:
            for s in schema.props:
                s_name = to_snack_case(s.name)
                s_typ, s_typ_kwargs, cls_builder_name = render_schema(s, cmd_ctx.response_clses, s_name)
                rendered_schemas.append((s_name, s_typ, s_typ_kwargs, cls_builder_name))
                if not cls_builder_name and isinstance(s, (CMDObjectSchemaBase, CMDArraySchemaBase)):
                    search_schemas[s_name] = s
        elif schema.additional_props:
            if schema.additional_props.item is not None:
                # AAZDictType
                s = schema.additional_props.item
                s_name = "Element"
                s_typ, s_typ_kwargs, cls_builder_name = render_schema_base(s, cmd_ctx.response_clses)
                rendered_schemas.append((s_name, s_typ, s_typ_kwargs, cls_builder_name))
                if not cls_builder_name and isinstance(s, (CMDObjectSchemaBase, CMDArraySchemaBase)):
                    search_schemas[s_name] = s
            else:
                assert schema.additional_props.any_type is True
        else:
            raise NotImplementedError()

    elif isinstance(schema, CMDObjectSchemaDiscriminator):
        if schema.discriminators:
            discriminators.extend(schema.discriminators)

        if schema.props:
            for s in schema.props:
                s_name = to_snack_case(s.name)
                s_typ, s_typ_kwargs, cls_builder_name = render_schema(s, cmd_ctx.response_clses, s_name)
                rendered_schemas.append((s_name, s_typ, s_typ_kwargs, cls_builder_name))
                if not cls_builder_name and isinstance(s, (CMDObjectSchemaBase, CMDArraySchemaBase)):
                    search_schemas[s_name] = s
        else:
            raise NotImplementedError()

    elif isinstance(schema, CMDArraySchemaBase):
        # AAZListType
        assert schema.item is not None
        s = schema.item
        s_name = "Element"
        s_typ, s_typ_kwargs, cls_builder_name = render_schema_base(s, cmd_ctx.response_clses)
        rendered_schemas.append((s_name, s_typ, s_typ_kwargs, cls_builder_name))
        if not cls_builder_name and isinstance(s, (CMDObjectSchemaBase, CMDArraySchemaBase)):
            search_schemas[s_name] = s

    else:
        raise NotImplementedError()

    if rendered_schemas:
        yield name, scope_define, rendered_schemas

    for s_name, s in search_schemas.items():
        s_scope_define = f"{scope_define}.{s_name}"
        if s_name == "Element":
            s_name = '_element'
        for scopes in _iter_response_scopes_by_schema_base(s, s_name, s_scope_define, cmd_ctx):
            yield scopes

    for disc in discriminators:
        key_name = to_snack_case(disc.property)
        key_value = disc.value
        disc_name = f"disc_{to_snack_case(disc.value)}"
        disc_scope_define = f'{scope_define}.discriminate_by("{key_name}", "{key_value}")'
        for scopes in _iter_response_scopes_by_schema_base(disc, disc_name, disc_scope_define, cmd_ctx):
            yield scopes


def render_schema(schema, cls_map, name):
    schema_kwargs = {}
    if name != schema.name:
        schema_kwargs['serialized_name'] = schema.name

    flags = {}

    if schema.required:
        flags['required'] = True
    if schema.skip_url_encoding:
        flags['skip_quote'] = True
    if getattr(schema, 'client_flatten', False):
        flags['client_flatten'] = True
    if getattr(schema, 'secret', False):
        flags['secret'] = True

    if flags:
        schema_kwargs['flags'] = flags

    schema_type, schema_kwargs, cls_builder_name = render_schema_base(schema, cls_map, schema_kwargs)

    return schema_type, schema_kwargs, cls_builder_name


def render_schema_base(schema, cls_map, schema_kwargs=None):
    if isinstance(schema, CMDClsSchemaBase):
        cls_name = schema.type[1:]
        schema = cls_map[cls_name].schema
    else:
        cls_name = getattr(schema, 'cls', None)
    cls_builder_name = parse_cls_builder_name(cls_name) if cls_name else None

    if schema_kwargs is None:
        schema_kwargs = {}

    if schema.nullable:
        schema_kwargs['nullable'] = True

    flags = schema_kwargs.get('flags', {})

    if schema.read_only:
        flags['read_only'] = True

    if isinstance(schema, CMDStringSchemaBase):
        schema_type = "AAZStrType"
    elif isinstance(schema, CMDIntegerSchemaBase):
        schema_type = "AAZIntType"
    elif isinstance(schema, CMDBooleanSchemaBase):
        schema_type = "AAZBoolType"
    elif isinstance(schema, CMDFloatSchemaBase):
        schema_type = "AAZFloatType"
    elif isinstance(schema, CMDObjectSchemaBase):
        if schema.props or schema.discriminators:
            schema_type = "AAZObjectType"
            if schema.additional_props:
                raise NotImplementedError()
        elif schema.additional_props:
            if schema.additional_props.any_type is True:
                schema_type = "AAZFreeFormDictType"
            else:
                assert schema.additional_props.item is not None
                schema_type = "AAZDictType"
        else:
            # empty object
            schema_type = "AAZObjectType"
    elif isinstance(schema, CMDArraySchemaBase):
        schema_type = "AAZListType"
    else:
        raise NotImplementedError()

    if flags:
        schema_kwargs['flags'] = flags

    return schema_type, schema_kwargs, cls_builder_name


def parse_cls_builder_name(cls_name):
    return f"_build_schema_{to_snack_case(cls_name)}"



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from schematics.models import Model
from schematics.types import StringType, ListType, ModelType, PolyModelType
from schematics.types.serializable import serializable

from ._schema import CMDSchemaField
from ._arg_builder import CMDArgBuilder


class CMDSelectorIndexBase(Model):
    TYPE_VALUE = None

    class Options:
        serialize_when_none = False

    @serializable
    def type(self):
        return self._get_type()

    def generate_args(self, ref_args, var_prefix):
        raise NotImplementedError()

    def reformat(self, **kwargs):
        raise NotImplementedError()

    def _get_type(self):
        assert self.TYPE_VALUE is not None
        return self.TYPE_VALUE

    @classmethod
    def _claim_polymorphic(cls, data):
        if cls.TYPE_VALUE is None:
            return False

        if isinstance(data, dict):
            type_value = data.get('type', None)
            if type_value is not None:
                typ = type_value.replace("<", " ").replace(">", " ").strip().split()[0]
                return typ == cls.TYPE_VALUE
        elif isinstance(data, CMDSelectorIndexBase):
            return data.TYPE_VALUE == cls.TYPE_VALUE
        return False


class CMDSelectorIndexBaseField(PolyModelType):

    def __init__(self, **kwargs):
        super(CMDSelectorIndexBaseField, self).__init__(
            model_spec=CMDSelectorIndexBase,
            allow_subclasses=True,
            serialize_when_none=False,
            **kwargs
        )

    def find_model(self, data):
        if self.claim_function:
            kls = self.claim_function(self, data)
            if not kls:
                raise Exception("Input for polymorphic field did not match any model")
            return kls

        fallback = None
        matching_classes = set()
        for kls in self._get_candidates():
            if issubclass(kls, CMDSelectorIndex):
                continue

            try:
                kls_claim = kls._claim_polymorphic
            except AttributeError:
                if not fallback:
                    fallback = kls
            else:
                if kls_claim(data):
                    matching_classes.add(kls)

        if not matching_classes and fallback:
            return fallback
        elif len(matching_classes) != 1:
            raise Exception("Got ambiguous input for polymorphic field")

        return matching_classes.pop()


class CMDSelectorIndex(CMDSelectorIndexBase):
    # properties as tags
    name = StringType(required=True)

    @classmethod
    def _claim_polymorphic(cls, data):
        if super(CMDSelectorIndex, cls)._claim_polymorphic(data):
            if isinstance(data, dict):
                # distinguish with CMDSchemaBase and CMDSchema
                return 'name' in data
            else:
                return isinstance(data, CMDSelectorIndex)
        return False


class CMDSelectorIndexField(PolyModelType):

    def __init__(self, **kwargs):
        super(CMDSelectorIndexField, self).__init__(
            model_spec=CMDSelectorIndex,
            allow_subclasses=True,
            serialize_when_none=False,
            **kwargs
        )

# object


class CMDObjectIndexDiscriminator(Model):
    # properties as tags
    property = StringType(required=True)
    value = StringType(required=True)

    # properties as nodes
    prop = CMDSelectorIndexField()
    discriminator = ModelType('CMDObjectIndexDiscriminator')

    class Options:
        serialize_when_none = False

    def generate_args(self, ref_args, var_prefix):
        if var_prefix.endswith("$"):
            var_prefix += f'{self.value}'
        else:
            var_prefix += f'.{self.value}'

        args = []
        if self.prop:
            args.extend(self.prop.generate_args(ref_args, var_prefix))
        if self.discriminator:
            args.extend(self.discriminator.generate_args(ref_args, var_prefix))

        return args

    def reformat(self, **kwargs):
        if self.prop:
            self.prop.reformat(**kwargs)
        if self.discriminator:
            self.discriminator.reformat(**kwargs)


class CMDObjectIndexAdditionalProperties(Model):
    item = CMDSelectorIndexBaseField(required=True)
    identifiers = ListType(CMDSchemaField())

    class Options:
        serialize_when_none = False

    def generate_args(self, ref_args, var_prefix):
        args = []
        if self.identifiers:
            for identifier in self.identifiers:
                builder = CMDArgBuilder.new_builder(schema=identifier, ref_args=ref_args, var_prefix=var_prefix)
                identifier_args = builder.get_args()
                assert len(identifier_args) == 1
                args.append(identifier_args[0])

        var_prefix += '{}'
        args.extend(self.item.generate_args(ref_args, var_prefix))
        return args

    def reformat(self, **kwargs):
        if self.item:
            self.item.reformat(**kwargs)
        if self.identifiers:
            for identifier in self.identifiers:
                identifier.reformat(**kwargs)
            self.identifiers = sorted(self.identifiers, key=lambda i: i.name)


class CMDObjectIndexBase(CMDSelectorIndexBase):
    TYPE_VALUE = "object"

    prop = CMDSelectorIndexField()
    discriminator = ModelType(CMDObjectIndexDiscriminator)
    additional_props = ModelType(
        CMDObjectIndexAdditionalProperties,
        serialized_name="additionalProps",
        deserialize_from="additionalProps",
    )

    def _generate_args_base(self, ref_args, var_prefix):
        args = []
        if self.prop:
            args.extend(self.prop.generate_args(ref_args, var_prefix))
        if self.discriminator:
            args.extend(self.discriminator.generate_args(ref_args, var_prefix))
        if self.additional_props:
            args.extend(self.additional_props.generate_args(ref_args, var_prefix))
        return args

    def generate_args(self, ref_args, var_prefix):
        return self._generate_args_base(ref_args, var_prefix)

    def reformat(self, **kwargs):
        if self.prop:
            self.prop.reformat(**kwargs)
        if self.discriminator:
            self.discriminator.reformat(**kwargs)
        if self.additional_props:
            self.additional_props.reformat(**kwargs)


class CMDObjectIndex(CMDObjectIndexBase, CMDSelectorIndex):

    def generate_args(self, ref_args, var_prefix):
        if var_prefix.endswith("$"):
            var_prefix += f'{self.name}'
        else:
            var_prefix += f'.{self.name}'

        return self._generate_args_base(ref_args, var_prefix)

# array


class CMDArrayIndexBase(CMDSelectorIndexBase):
    TYPE_VALUE = "array"

    item = CMDSelectorIndexBaseField()
    identifiers = ListType(CMDSchemaField())

    def _generate_args_base(self, ref_args, var_prefix):
        args = []
        if self.identifiers:
            for identifier in self.identifiers:
                builder = CMDArgBuilder.new_builder(schema=identifier, ref_args=ref_args, var_prefix=var_prefix)
                identifier_args = builder.get_args()
                assert len(identifier_args) == 1
                args.append(identifier_args[0])

        var_prefix += "[]"
        if self.item:
            args.extend(self.item.generate_args(ref_args, var_prefix))

        return args

    def generate_args(self, ref_args, var_prefix):
        return self._generate_args_base(ref_args, var_prefix)

    def reformat(self, **kwargs):
        if self.item:
            self.item.reformat(**kwargs)
        if self.identifiers:
            for identifier in self.identifiers:
                identifier.reformat(**kwargs)
            self.identifiers = sorted(self.identifiers, key=lambda i: i.name)


class CMDArrayIndex(CMDArrayIndexBase, CMDSelectorIndex):

    def generate_args(self, ref_args, var_prefix):
        if var_prefix.endswith("$"):
            var_prefix += f'{self.name}'
        else:
            var_prefix += f'.{self.name}'

        return self._generate_args_base(ref_args, var_prefix)



import json

from flask import Flask, jsonify
from flask.cli import FlaskGroup, shell_command, routes_command
from flask.logging import create_logger

from utils import exceptions
from utils.config import Config
from aaz_dev.app.run import run_command


def create_app():
    app = Flask(__name__, static_folder=Config.STATIC_FOLDER, static_url_path=Config.STATIC_URL_PATH)
    logger = create_logger(app)

    @app.route('/', defaults={'path': ''})
    @app.route('/<string:path>')
    @app.route('/<path:path>')
    def index(path):
        if '.' in path:
            return app.send_static_file(path)
        return app.send_static_file('index.html')

    @app.errorhandler(exceptions.InvalidAPIUsage)
    def invalid_api_usage(e):
        logger.error(f'InvalidAPIUsage:\n{json.dumps(e.to_dict(), indent=2)}',)
        return jsonify(e.to_dict()), e.status_code

    # register url converters
    from .url_converters import Base64Converter, NameConverter, NameWithCapitalConverter, NamesPathConverter, ListPathConvertor
    app.url_map.converters['base64'] = Base64Converter
    app.url_map.converters['name'] = NameConverter
    app.url_map.converters['Name'] = NameWithCapitalConverter
    app.url_map.converters['names_path'] = NamesPathConverter
    app.url_map.converters['list_path'] = ListPathConvertor

    # register routes of swagger module
    from swagger.api import register_blueprints
    register_blueprints(app)

    # register routes of command module
    from command.api import register_blueprints
    register_blueprints(app)

    # register routes of cli module
    from cli.api import register_blueprints
    register_blueprints(app)

    return app


cli = FlaskGroup(
    create_app=create_app,
    add_default_commands=False
)

cli.add_command(shell_command)
cli.add_command(routes_command)
cli.add_command(run_command)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from schematics.models import Model
from schematics.types import ModelType, ListType, PolyModelType
from schematics.types.serializable import serializable

from ._arg import CMDStringArg, CMDStringArgBase, \
    CMDByteArg, CMDByteArgBase, \
    CMDBinaryArg, CMDBinaryArgBase, \
    CMDDurationArg, CMDDurationArgBase, \
    CMDDateArg, CMDDateArgBase, \
    CMDDateTimeArg, CMDDateTimeArgBase, \
    CMDUuidArg, CMDUuidArgBase, \
    CMDPasswordArg, CMDPasswordArgBase, \
    CMDResourceIdArg, CMDResourceIdArgBase, \
    CMDResourceLocationArg, CMDResourceLocationArgBase, \
    CMDBooleanArg, CMDBooleanArgBase, \
    CMDIntegerArg, CMDIntegerArgBase, \
    CMDInteger32Arg, CMDInteger32ArgBase, \
    CMDInteger64Arg, CMDInteger64ArgBase, \
    CMDFloatArg, CMDFloatArgBase, \
    CMDFloat32Arg, CMDFloat32ArgBase, \
    CMDFloat64Arg, CMDFloat64ArgBase, \
    CMDArrayArg, CMDArrayArgBase, \
    CMDObjectArg, CMDObjectArgBase, CMDObjectArgAdditionalProperties, \
    CMDClsArg, CMDClsArgBase
from ._fields import CMDVariantField, StringType, CMDClassField, CMDBooleanField, CMDPrimitiveField, CMDDescriptionField
from ._format import CMDStringFormat, CMDIntegerFormat, CMDFloatFormat, CMDObjectFormat, CMDArrayFormat, \
    CMDResourceIdFormat
from ._utils import CMDDiffLevelEnum
from utils import exceptions

import logging


logger = logging.getLogger('backend')

class CMDSchemaEnumItem(Model):
    arg = CMDVariantField()  # value will be used when specific argument is provided

    # properties as nodes
    value = CMDPrimitiveField()  # json value format string, support null

    class Options:
        serialize_when_none = False

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.value != old.value:
                diff["value"] = f"{old.value} != {self.value}"

        if level >= CMDDiffLevelEnum.Associate:
            if self.arg != old.arg:
                diff["arg"] = f"{old.arg} != {self.arg}"

        return diff


class CMDSchemaEnum(Model):
    # properties as tags

    # properties as nodes
    items = ListType(ModelType(CMDSchemaEnumItem), min_size=1)

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = []

        if level >= CMDDiffLevelEnum.BreakingChange:
            # old.items should be the subset of self.items
            for old_item in old.items:
                matched = False
                for item in self.items:
                    if old_item.value == item.value:
                        matched = True
                        item_diff = item.diff(old_item, level)
                        if item_diff:
                            diff.append(item_diff)
                        break
                if not matched:
                    diff.append(f"MissEnumItem: {old_item.value}")

        if level >= CMDDiffLevelEnum.Structure:
            for item in self.items:
                matched = False
                for old_item in old.items:
                    if item.value == old_item.value:
                        matched = True
                        break
                if not matched:
                    diff.append(f"NewEnumItem: {item.value}")
        return diff

    def reformat(self, **kwargs):
        try:
            self.items = sorted(self.items, key=lambda it: it.value)
        except:
            # some of the value main not support sort
            pass


class CMDSchemaDefault(Model):
    """ The argument value if an argument is not used """

    # properties as nodes
    value = CMDPrimitiveField()  # json value format string, support null

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.value != old.value:
                diff["value"] = f"{old.value} != {self.value}"
        return diff


class CMDSchemaBase(Model):
    TYPE_VALUE = None
    ARG_TYPE = None

    # properties as tags
    read_only = CMDBooleanField(
        serialized_name="readOnly",
        deserialize_from="readOnly"
    )
    frozen = CMDBooleanField()  # frozen schema will not be used
    const = CMDBooleanField()  # when a schema is const it's default value is not None.

    # properties as nodes
    default = ModelType(CMDSchemaDefault)
    nullable = CMDBooleanField()  # whether null value is supported

    # base types: "array", "boolean", "integer", "float", "object", "string",

    class Options:
        serialize_when_none = False

    @serializable
    def type(self):
        return self._get_type()

    def _get_type(self):
        assert self.TYPE_VALUE is not None
        return self.TYPE_VALUE

    @classmethod
    def _claim_polymorphic(cls, data):
        if cls.TYPE_VALUE is None:
            return False

        if isinstance(data, dict):
            type_value = data.get('type', None)
            if type_value is not None:
                typ = type_value.replace("<", " ").replace(">", " ").strip().split()[0]
                return typ == cls.TYPE_VALUE
        elif isinstance(data, CMDSchemaBase):
            return data.TYPE_VALUE == cls.TYPE_VALUE
        return False

    def _diff_base(self, old, level, diff):
        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.type != old.type:
                diff["type"] = f"{old.type} != {self.type}"
            if self.read_only and not old.read_only:
                diff["read_only"] = f"it's read_only now."
            if self.const and not old.const:
                diff["const"] = f"it's const now."
            if old.default:
                if not self.default:
                    diff["default"] = f"miss default now."
                else:
                    default_diff = self.default.diff(old.default, level)
                    if default_diff:
                        diff["default"] = default_diff

        if level >= CMDDiffLevelEnum.Structure:
            if (not self.read_only) != (not old.read_only):
                diff["read_only"] = f"{old.read_only} != {self.read_only}"
            if (not self.const) != (not old.const):
                diff['const'] = f"{old.const} != {self.const}"
            if self.default:
                default_diff = self.default.diff(old.default, level)
                if default_diff:
                    diff["default"] = default_diff
        return diff

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        if self.frozen and old.frozen:
            return None
        diff = {}
        diff = self._diff_base(old, level, diff)
        return diff

    def _reformat_base(self, **kwargs):
        pass

    def reformat(self, **kwargs):
        if self.frozen:
            return
        self._reformat_base(**kwargs)


class CMDSchemaBaseField(PolyModelType):

    def __init__(self, **kwargs):
        super(CMDSchemaBaseField, self).__init__(
            model_spec=CMDSchemaBase,
            allow_subclasses=True,
            serialize_when_none=False,
            **kwargs
        )

    def export(self, value, format, context=None):
        if value.frozen:
            # frozen schema base will be ignored
            return None
        return super(CMDSchemaBaseField, self).export(value, format, context)

    def find_model(self, data):
        if self.claim_function:
            kls = self.claim_function(self, data)
            if not kls:
                raise Exception("Input for polymorphic field did not match any model")
            return kls

        fallback = None
        matching_classes = set()
        for kls in self._get_candidates():
            if issubclass(kls, CMDSchema):
                continue

            try:
                kls_claim = kls._claim_polymorphic
            except AttributeError:
                if not fallback:
                    fallback = kls
            else:
                if kls_claim(data):
                    matching_classes.add(kls)

        if not matching_classes and fallback:
            return fallback
        elif len(matching_classes) != 1:
            raise Exception("Got ambiguous input for polymorphic field")

        return matching_classes.pop()


class CMDSchema(CMDSchemaBase):
    # properties as tags
    name = StringType(required=True)
    arg = CMDVariantField()
    required = CMDBooleanField()

    description = CMDDescriptionField()

    skip_url_encoding = CMDBooleanField(
        serialized_name="skipUrlEncoding",
        deserialize_from="skipUrlEncoding",
    )  # used in path and query parameters

    secret = CMDBooleanField()

    @classmethod
    def _claim_polymorphic(cls, data):
        if super(CMDSchema, cls)._claim_polymorphic(data):
            if isinstance(data, dict):
                # distinguish with CMDSchemaBase and CMDSchema
                return 'name' in data
            else:
                return isinstance(data, CMDSchema)
        return False

    def _diff(self, old, level, diff):
        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.name != old.name:
                diff["name"] = f"{old.name} != {self.name}"
            if self.required and not old.required:
                diff["required"] = f"it's required now."
            if (not self.skip_url_encoding) != (not old.skip_url_encoding):  # None should be same as false
                diff["skip_url_encoding"] = f"{old.skip_url_encoding} != {self.skip_url_encoding}"
            if not self.secret and old.secret:
                diff["secret"] = f"it's not secret property now."

        if level >= CMDDiffLevelEnum.Structure:
            if (not self.required) != (not old.required):
                diff["required"] = f"{old.required} != {self.required}"
            if self.secret != old.secret:
                diff["secret"] = f"{old.secret} != {self.secret}"

        if level >= CMDDiffLevelEnum.Associate:
            if self.arg != old.arg:
                diff["arg"] = f"{old.arg} != {self.arg}"

        if level >= CMDDiffLevelEnum.All:
            if self.description != old.description:
                diff["description"] = f"'{old.description}' != '{self.description}'"
        return diff

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        if self.frozen and old.frozen:
            return None
        diff = {}
        diff = self._diff_base(old, level, diff)
        diff = self._diff(old, level, diff)
        return diff

    def _reformat(self, **kwargs):
        pass

    def reformat(self, **kwargs):
        if self.frozen:
            return
        self._reformat_base(**kwargs)
        self._reformat(**kwargs)


class CMDSchemaField(PolyModelType):

    def __init__(self, **kwargs):
        super(CMDSchemaField, self).__init__(
            model_spec=CMDSchema,
            allow_subclasses=True,
            serialize_when_none=False,
            **kwargs
        )

    def export(self, value, format, context=None):
        if value.frozen:
            # frozen schema base will be ignored
            return None
        return super(CMDSchemaField, self).export(value, format, context)


# cls
class CMDClsSchemaBase(CMDSchemaBase):
    ARG_TYPE = CMDClsArgBase

    _type = StringType(
        deserialize_from='type',
        serialized_name='type',
        required=True
    )

    def _get_type(self):
        return self._type

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.implement = None

    @classmethod
    def _claim_polymorphic(cls, data):
        if isinstance(data, dict):
            type_value = data.get('type', None)
            if type_value is not None and type_value.startswith("@"):
                return True
        elif isinstance(data, CMDClsSchemaBase):
            return True
        return False

    @classmethod
    def build_from_schema_base(cls, schema_base, implement):
        assert isinstance(schema_base, (CMDObjectSchemaBase, CMDArraySchemaBase))
        assert getattr(schema_base, 'cls', None)
        cls_schema = cls()
        cls_schema._type = f"@{schema_base.cls}"
        cls_schema.read_only = schema_base.read_only
        cls_schema.frozen = schema_base.frozen
        cls_schema.const = schema_base.const
        cls_schema.default = schema_base.default
        cls_schema.implement = implement
        return cls_schema

    def get_unwrapped(self, **kwargs):
        if self.implement is None:
            return
        assert isinstance(self.implement, CMDSchemaBase)
        if isinstance(self.implement, CMDObjectSchemaBase):
            cls = CMDObjectSchema if isinstance(self, CMDClsSchema) else CMDObjectSchemaBase
        elif isinstance(self.implement, CMDArraySchemaBase):
            cls = CMDArraySchema if isinstance(self, CMDClsSchema) else CMDArraySchemaBase
        else:
            raise NotImplementedError()
        data = {k: v for k, v in self.implement.to_native().items() if k in cls._schema.valid_input_keys}
        data.update(kwargs)
        unwrapped = cls(data)
        return unwrapped


class CMDClsSchema(CMDClsSchemaBase, CMDSchema):
    ARG_TYPE = CMDClsArg

    # properties as tags
    client_flatten = CMDBooleanField(
        serialized_name="clientFlatten",
        deserialize_from="clientFlatten"
    )

    def _diff(self, old, level, diff):
        # TODO: Handle Cls Schema compare with other Schema classes
        diff = super(CMDClsSchema, self)._diff(old, level, diff)
        if level >= CMDDiffLevelEnum.BreakingChange:
            if (not self.client_flatten) != (not old.client_flatten):
                diff["client_flatten"] = f"from {old.client_flatten} to {self.client_flatten}"

        return diff

    @classmethod
    def build_from_schema(cls, schema, implement):
        assert isinstance(schema, (CMDObjectSchema, CMDArraySchema))
        cls_schema = cls.build_from_schema_base(schema, implement)
        cls_schema.name = schema.name
        cls_schema.arg = schema.arg
        cls_schema.required = schema.required
        cls_schema.description = schema.description
        cls_schema.skip_url_encoding = schema.skip_url_encoding
        if isinstance(schema, CMDObjectSchema):
            cls_schema.client_flatten = schema.client_flatten
        return cls_schema

    def get_unwrapped(self, **kwargs):
        uninherent = {
            "name": self.name,
            "arg": self.arg,
            "required": self.required,
            "description": self.description,
            "skip_url_encoding": self.skip_url_encoding,
        }
        if isinstance(self.implement, CMDObjectSchema):
            uninherent["client_flatten"] = self.client_flatten
        uninherent.update(kwargs)
        return super().get_unwrapped(**uninherent)


# string
class CMDStringSchemaBase(CMDSchemaBase):
    TYPE_VALUE = "string"
    ARG_TYPE = CMDStringArgBase

    fmt = ModelType(
        CMDStringFormat,
        serialized_name='format',
        deserialize_from='format'
    )
    enum = ModelType(CMDSchemaEnum)

    def _diff_base(self, old, level, diff):
        diff = super(CMDStringSchemaBase, self)._diff_base(old, level, diff)

        fmt_diff = _diff_fmt(self.fmt, old.fmt, level)
        if fmt_diff:
            diff["fmt"] = fmt_diff

        enum_diff = _diff_enum(self.enum, old.enum, level)
        if enum_diff:
            diff["enum"] = enum_diff

        return diff

    def _reformat_base(self, **kwargs):
        super()._reformat_base(**kwargs)
        if self.enum:
            self.enum.reformat(**kwargs)


class CMDStringSchema(CMDStringSchemaBase, CMDSchema):
    ARG_TYPE = CMDStringArg


# byte: base64 encoded characters
class CMDByteSchemaBase(CMDStringSchemaBase):
    TYPE_VALUE = "byte"
    ARG_TYPE = CMDByteArgBase


class CMDByteSchema(CMDByteSchemaBase, CMDStringSchema):
    ARG_TYPE = CMDByteArg


# binary: any sequence of octets
class CMDBinarySchemaBase(CMDStringSchemaBase):
    TYPE_VALUE = "binary"
    ARG_TYPE = CMDBinaryArgBase


class CMDBinarySchema(CMDBinarySchemaBase, CMDStringSchema):
    ARG_TYPE = CMDBinaryArg


# duration
class CMDDurationSchemaBase(CMDStringSchemaBase):
    TYPE_VALUE = "duration"
    ARG_TYPE = CMDDurationArgBase


class CMDDurationSchema(CMDDurationSchemaBase, CMDStringSchema):
    ARG_TYPE = CMDDurationArg


# date: As defined by full-date - https://xml2rfc.tools.ietf.org/public/rfc/html/rfc3339.html#anchor14
class CMDDateSchemaBase(CMDStringSchemaBase):
    TYPE_VALUE = "date"
    ARG_TYPE = CMDDateArgBase


class CMDDateSchema(CMDDateSchemaBase, CMDStringSchema):
    ARG_TYPE = CMDDateArg


# date-time: As defined by date-time - https://xml2rfc.tools.ietf.org/public/rfc/html/rfc3339.html#anchor14
class CMDDateTimeSchemaBase(CMDStringSchemaBase):
    TYPE_VALUE = "dateTime"
    ARG_TYPE = CMDDateTimeArgBase


class CMDDateTimeSchema(CMDDateTimeSchemaBase, CMDStringSchema):
    ARG_TYPE = CMDDateTimeArg


# uuid
class CMDUuidSchemaBase(CMDStringSchemaBase):
    TYPE_VALUE = "uuid"
    ARG_TYPE = CMDUuidArgBase


class CMDUuidSchema(CMDUuidSchemaBase, CMDStringSchema):
    ARG_TYPE = CMDUuidArg


# password
class CMDPasswordSchemaBase(CMDStringSchemaBase):
    TYPE_VALUE = "password"
    ARG_TYPE = CMDPasswordArgBase


class CMDPasswordSchema(CMDPasswordSchemaBase, CMDStringSchema):
    ARG_TYPE = CMDPasswordArg


# ResourceId
class CMDResourceIdSchemaBase(CMDStringSchemaBase):
    TYPE_VALUE = "ResourceId"
    ARG_TYPE = CMDResourceIdArgBase

    fmt = ModelType(
        CMDResourceIdFormat,
        serialized_name='format',
        deserialize_from='format'
    )


class CMDResourceIdSchema(CMDResourceIdSchemaBase, CMDStringSchema):
    ARG_TYPE = CMDResourceIdArg


# ResourceLocation
class CMDResourceLocationSchemaBase(CMDStringSchemaBase):
    TYPE_VALUE = "ResourceLocation"
    ARG_TYPE = CMDResourceLocationArgBase


class CMDResourceLocationSchema(CMDResourceLocationSchemaBase, CMDStringSchema):
    ARG_TYPE = CMDResourceLocationArg


# integer
class CMDIntegerSchemaBase(CMDSchemaBase):
    TYPE_VALUE = "integer"
    ARG_TYPE = CMDIntegerArgBase

    fmt = ModelType(
        CMDIntegerFormat,
        serialized_name='format',
        deserialize_from='format',
    )
    enum = ModelType(CMDSchemaEnum)

    def _diff_base(self, old, level, diff):
        diff = super(CMDIntegerSchemaBase, self)._diff_base(old, level, diff)

        fmt_diff = _diff_fmt(self.fmt, old.fmt, level)
        if fmt_diff:
            diff["fmt"] = fmt_diff

        enum_diff = _diff_enum(self.enum, old.enum, level)
        if enum_diff:
            diff["enum"] = enum_diff

        return diff

    def _reformat_base(self, **kwargs):
        super()._reformat_base(**kwargs)
        if self.enum:
            self.enum.reformat(**kwargs)


class CMDIntegerSchema(CMDIntegerSchemaBase, CMDSchema):
    ARG_TYPE = CMDIntegerArg


# integer32
class CMDInteger32SchemaBase(CMDIntegerSchemaBase):
    TYPE_VALUE = "integer32"
    ARG_TYPE = CMDInteger32ArgBase


class CMDInteger32Schema(CMDInteger32SchemaBase, CMDIntegerSchema):
    ARG_TYPE = CMDInteger32Arg


# integer64
class CMDInteger64SchemaBase(CMDIntegerSchemaBase):
    TYPE_VALUE = "integer64"
    ARG_TYPE = CMDInteger64ArgBase


class CMDInteger64Schema(CMDInteger64SchemaBase, CMDIntegerSchema):
    ARG_TYPE = CMDInteger64Arg


# boolean
class CMDBooleanSchemaBase(CMDSchemaBase):
    TYPE_VALUE = "boolean"
    ARG_TYPE = CMDBooleanArgBase


class CMDBooleanSchema(CMDBooleanSchemaBase, CMDSchema):
    ARG_TYPE = CMDBooleanArg


# float
class CMDFloatSchemaBase(CMDSchemaBase):
    TYPE_VALUE = "float"
    ARG_TYPE = CMDFloatArgBase

    fmt = ModelType(
        CMDFloatFormat,
        serialized_name='format',
        deserialize_from='format',
    )
    enum = ModelType(CMDSchemaEnum)

    def _diff_base(self, old, level, diff):
        diff = super(CMDFloatSchemaBase, self)._diff_base(old, level, diff)

        fmt_diff = _diff_fmt(self.fmt, old.fmt, level)
        if fmt_diff:
            diff["fmt"] = fmt_diff

        enum_diff = _diff_enum(self.enum, old.enum, level)
        if enum_diff:
            diff["enum"] = enum_diff

        return diff

    def _reformat_base(self, **kwargs):
        super()._reformat_base(**kwargs)
        if self.enum:
            self.enum.reformat(**kwargs)


class CMDFloatSchema(CMDFloatSchemaBase, CMDSchema):
    ARG_TYPE = CMDFloatArg


# float32
class CMDFloat32SchemaBase(CMDFloatSchemaBase):
    TYPE_VALUE = "float32"
    ARG_TYPE = CMDFloat32ArgBase


class CMDFloat32Schema(CMDFloat32SchemaBase, CMDFloatSchema):
    ARG_TYPE = CMDFloat32Arg


# float64
class CMDFloat64SchemaBase(CMDFloatSchemaBase):
    TYPE_VALUE = "float64"
    ARG_TYPE = CMDFloat64ArgBase


class CMDFloat64Schema(CMDFloat64SchemaBase, CMDFloatSchema):
    ARG_TYPE = CMDFloat64Arg


# object

# discriminator

class CMDObjectSchemaDiscriminatorField(ModelType):

    def __init__(self, model_spec=None, **kwargs):
        super(CMDObjectSchemaDiscriminatorField, self).__init__(
            model_spec=model_spec or CMDObjectSchemaDiscriminator,
            serialize_when_none=False,
            **kwargs
        )

    def export(self, value, format, context=None):
        if hasattr(value, 'frozen') and value.frozen:
            # frozen schema base will be ignored
            return None
        return super(CMDObjectSchemaDiscriminatorField, self).export(value, format, context)


class CMDObjectSchemaDiscriminator(Model):
    ARG_TYPE = CMDObjectArg

    # properties as tags
    property = StringType(required=True)
    value = StringType(required=True)
    frozen = CMDBooleanField()  # frozen schema will not be used

    # properties as nodes
    props = ListType(CMDSchemaField())
    discriminators = ListType(CMDObjectSchemaDiscriminatorField(model_spec='CMDObjectSchemaDiscriminator'))

    class Options:
        serialize_when_none = False

    def diff(self, old, level):
        if self.frozen and old.frozen:
            return None
        diff = {}

        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.property != old.property:
                diff["property"] = f"{old.property} != {self.property}"
            if self.value != old.value:
                diff["value"] = f"{old.value} != {self.value}"

        props_diff = _diff_props(
            self.props or [],
            old.props or [],
            level
        )
        if props_diff:
            diff["props"] = props_diff

        discs_diff = _diff_discriminators(
            self.discriminators or [],
            old.discriminators or [],
            level
        )
        if discs_diff:
            diff["discriminators"] = discs_diff

        return diff

    def reformat(self, **kwargs):
        if self.frozen:
            return
        if self.props:
            for prop in self.props:
                prop.reformat(**kwargs)
            self.props = sorted(self.props, key=lambda p: p.name)
        if self.discriminators:
            for disc in self.discriminators:
                disc.reformat(**kwargs)
            self.discriminators = sorted(self.discriminators, key=lambda disc: disc.value)


# additionalProperties
class CMDObjectSchemaAdditionalProperties(Model):
    ARG_TYPE = CMDObjectArgAdditionalProperties

    # properties as tags
    read_only = CMDBooleanField(
        serialized_name="readOnly",
        deserialize_from="readOnly"
    )
    frozen = CMDBooleanField()

    # properties as nodes
    item = CMDSchemaBaseField()
    any_type = CMDBooleanField(
        serialized_name="anyType",
        deserialize_from="anyType"
    )

    def diff(self, old, level):
        if self.frozen and old.frozen:
            return None
        diff = {}

        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.read_only and not old.read_only:
                diff["read_only"] = f"it's read_only now."
            if not self.any_type and old.any_type:
                diff["any_type"] = f"it's not any_type now"

        if level >= CMDDiffLevelEnum.Structure:
            if self.any_type:
                if not old.any_type:
                    diff["any_type"] = f"Now support any_type"

        item_diff = _diff_item(self.item, old.item, level)
        if item_diff:
            diff["item"] = item_diff

        return diff

    def reformat(self, **kwargs):
        if self.frozen:
            return
        if self.item:
            if self.any_type:
                raise exceptions.VerificationError(
                    "InvalidAdditionalPropertiesDefinition",
                    details="Additional property defined 'item' and 'any_type'."
                )
            self.item.reformat(**kwargs)


class CMDObjectSchemaAdditionalPropertiesField(ModelType):

    def __init__(self, **kwargs):
        super(CMDObjectSchemaAdditionalPropertiesField, self).__init__(
            model_spec=CMDObjectSchemaAdditionalProperties,
            serialized_name="additionalProps",
            deserialize_from="additionalProps",
            serialize_when_none=False,
            **kwargs
        )

    def export(self, value, format, context=None):
        if value.frozen:
            return None
        return super(CMDObjectSchemaAdditionalPropertiesField, self).export(value, format, context)


class CMDObjectSchemaBase(CMDSchemaBase):
    TYPE_VALUE = "object"
    ARG_TYPE = CMDObjectArgBase

    fmt = ModelType(
        CMDObjectFormat,
        serialized_name='format',
        deserialize_from='format',
    )
    props = ListType(CMDSchemaField())
    discriminators = ListType(CMDObjectSchemaDiscriminatorField())
    additional_props = CMDObjectSchemaAdditionalPropertiesField()

    # define a schema cls which can be used by others,
    # cls definition will not include properties in CMDSchema only, such as following properties:
    #  - name
    #  - arg
    #  - required
    #  - description
    #  - skip_url_encoding
    #  - client_flatten
    cls = CMDClassField()

    def _diff_base(self, old, level, diff):
        diff = super(CMDObjectSchemaBase, self)._diff_base(old, level, diff)

        fmt_diff = _diff_fmt(self.fmt, old.fmt, level)
        if fmt_diff:
            diff["fmt"] = fmt_diff

        props_diff = _diff_props(
            self.props or [],
            old.props or [],
            level
        )
        if props_diff:
            diff["props"] = props_diff

        discs_diff = _diff_discriminators(
            self.discriminators or [],
            old.discriminators or [],
            level
        )
        if discs_diff:
            diff["discriminators"] = discs_diff

        if level >= CMDDiffLevelEnum.BreakingChange:
            if old.additional_props:
                if not self.additional_props:
                    diff["additional_props"] = f"Miss additional props"

        if level >= CMDDiffLevelEnum.Structure:
            if self.additional_props:
                if not old.additional_props:
                    diff["additional_props"] = f"New additional props"

        if self.additional_props and old.additional_props:
            additional_diff = self.additional_props.diff(old.additional_props, level)
            if additional_diff:
                diff["additional_props"] = additional_diff

        return diff

    def _reformat_base(self, **kwargs):
        super()._reformat_base(**kwargs)
        if self.props:
            for prop in self.props:
                prop.reformat(**kwargs)
            self.props = sorted(self.props, key=lambda prop: prop.name)
        if self.discriminators:
            for disc in self.discriminators:
                disc.reformat(**kwargs)
            self.discriminators = sorted(self.discriminators, key=lambda disc: disc.value)
        if self.additional_props:
            self.additional_props.reformat(**kwargs)


class CMDObjectSchema(CMDObjectSchemaBase, CMDSchema):
    ARG_TYPE = CMDObjectArg

    # properties as tags
    client_flatten = CMDBooleanField(
        serialized_name="clientFlatten",
        deserialize_from="clientFlatten"
    )

    def _diff(self, old, level, diff):
        diff = super(CMDObjectSchema, self)._diff(old, level, diff)
        if level >= CMDDiffLevelEnum.BreakingChange:
            if (not self.client_flatten) != (not old.client_flatten):
                diff["client_flatten"] = f"from {old.client_flatten} to {self.client_flatten}"

        cls_diff = _diff_cls(self.cls, old.cls, level)
        if cls_diff:
            diff["cls"] = cls_diff

        return diff


class CMDIdentityObjectSchemaBase(CMDObjectSchemaBase):
    """ And identity object which contains 'userAssignedIdentities' property and 'type' property
    with "SystemAssigned", "UserAssigned", "SystemAssigned, UserAssigned" and "None" enum values.
    """
    TYPE_VALUE = "IdentityObject"
    ARG_TYPE = CMDObjectArgBase


class CMDIdentityObjectSchema(CMDIdentityObjectSchemaBase, CMDObjectSchema):
    ARG_TYPE = CMDObjectArg


# array
class CMDArraySchemaBase(CMDSchemaBase):
    TYPE_VALUE = "array"
    ARG_TYPE = CMDArrayArgBase

    # properties as nodes
    fmt = ModelType(
        CMDArrayFormat,
        serialized_name='format',
        deserialize_from='format',
    )
    item = CMDSchemaBaseField()

    # used to indentify item in array
    identifiers = ListType(StringType())

    # properties as tags
    # define a schema which can be used by others
    # cls definition will not include properties in CMDSchema only, such as following properties:
    #  - name
    #  - arg
    #  - required
    #  - description
    #  - skip_url_encoding
    cls = CMDClassField()

    def _get_type(self):
        return f"{self.TYPE_VALUE}<{self.item.type}>"

    def _diff_base(self, old, level, diff):
        diff = super(CMDArraySchemaBase, self)._diff_base(old, level, diff)

        fmt_diff = _diff_fmt(self.fmt, old.fmt, level)
        if fmt_diff:
            diff["fmt"] = fmt_diff

        item_diff = _diff_item(self.item, old.item, level)
        if item_diff:
            diff["item"] = item_diff

        if level >= CMDDiffLevelEnum.Structure:
            if old.identifiers:
                if not self.identifiers:
                    diff["identifiers"] = f"Miss Identifiers"
                elif set(old.identifiers) != set(self.identifiers):
                    diff["identifiers"] = f"Identifier different"
            elif self.identifiers:
                diff["identifiers"] = f"New identifiers"

        return diff

    def _reformat_base(self, **kwargs):
        super()._reformat_base(**kwargs)
        self.item.reformat(**kwargs)
        if self.identifiers:
            identifiers = sorted(self.identifiers, key=lambda i: (len(i), i))
            item_instance = self.item
            if isinstance(item_instance, CMDClsSchemaBase):
                item_instance = item_instance.implement
            if not isinstance(item_instance, CMDObjectSchemaBase):
                raise exceptions.InvalidAPIUsage(
                    f"Identifiers should be used in 'array of object'"
                )
            item_prop_names = {p.name for p in item_instance.props}
            new_identifiers = []
            for identifier in identifiers:
                if identifier not in item_prop_names:
                    if '/' in identifier:
                        # not support identifier with '/' right now
                        logger.info(f"identifier '{identifier}' is not supported yet")
                        continue
                    raise exceptions.InvalidAPIUsage(
                        f"identifier property '{identifier}' not exist"
                    )
                new_identifiers.append(identifier)
            self.identifiers = new_identifiers or None


class CMDArraySchema(CMDArraySchemaBase, CMDSchema):
    ARG_TYPE = CMDArrayArg

    def _diff(self, old, level, diff):
        diff = super(CMDArraySchema, self)._diff(old, level, diff)

        cls_diff = _diff_cls(self.cls, old.cls, level)
        if cls_diff:
            diff["cls"] = cls_diff

        return diff


# diff functions

def _diff_fmt(self_fmt, old_fmt, level):
    fmt_diff = None

    if level >= CMDDiffLevelEnum.Structure:
        if old_fmt:
            if not self_fmt:
                fmt_diff = f"Miss property"

    if self_fmt and old_fmt:
        fmt_diff = self_fmt.diff(old_fmt, level)

    return fmt_diff


def _diff_enum(self_enum, old_enum, level):
    enum_diff = None

    if level >= CMDDiffLevelEnum.Structure:
        if old_enum:
            if not self_enum:
                enum_diff = f"Miss property"

    if self_enum and old_enum:
        enum_diff = self_enum.diff(old_enum, level)

    return enum_diff


def _diff_props(self_props, old_props, level):
    props_diff = {}
    if level >= CMDDiffLevelEnum.BreakingChange:
        props_dict = {prop.name: prop for prop in self_props}
        for old_prop in old_props:
            if old_prop.name not in props_dict:
                if not old_prop.frozen:
                    props_diff[old_prop.name] = "Miss property"
            else:
                prop = props_dict.pop(old_prop.name)
                diff = prop.diff(old_prop, level)
                if diff:
                    props_diff[old_prop.name] = diff

        for prop in props_dict.values():
            if prop.frozen:
                continue
            if prop.required:
                props_diff[prop.name] = "New required property"

    if level >= CMDDiffLevelEnum.Structure:
        old_props_dict = {prop.name: prop for prop in old_props}
        for prop in self_props:
            if prop.name not in old_props_dict:
                if not prop.frozen:
                    props_diff[prop.name] = "New property"

    return props_diff


def _diff_discriminators(self_discriminators, old_discriminators, level):
    discs_diff = {}
    if level >= CMDDiffLevelEnum.BreakingChange:
        discs_dict = {disc.value: disc for disc in self_discriminators}
        for old_disc in old_discriminators:
            if old_disc.value not in discs_dict:
                if not old_disc.frozen:
                    discs_diff[old_disc.value] = "Miss discriminator value"
            else:
                disc = discs_dict.pop(old_disc.value)
                diff = disc.diff(old_disc, level)
                if diff:
                    discs_diff[old_disc.value] = diff

    if level >= CMDDiffLevelEnum.Structure:
        old_discs_dict = {disc.value: disc for disc in old_discriminators}
        for disc in self_discriminators:
            if disc.value not in old_discs_dict:
                if not disc.frozen:
                    discs_diff[disc.value] = "New discriminator value"

    return discs_diff


def _diff_item(self_item, old_item, level):
    item_diff = {}

    if self_item is None and old_item is None:
        return item_diff

    if level >= CMDDiffLevelEnum.BreakingChange:
        if type(self_item) is not type(old_item):
            item_diff = f"Type: {type(old_item)} != {type(self_item)}"
        elif not (self_item.frozen and old_item.frozen):
            item_diff = {}
            item_diff = self_item._diff_base(old_item, level, item_diff)

    return item_diff


def _diff_cls(self_cls, old_cls, level):
    cls_diff = None

    if level >= CMDDiffLevelEnum.Structure:
        if (self_cls is not None) != (old_cls is not None):
            cls_diff = f"from {old_cls} to {self_cls}"

    if level >= CMDDiffLevelEnum.All:
        if self_cls != old_cls:
            cls_diff = f"{old_cls} != {self_cls}"
    return cls_diff



from schematics.models import Model
from schematics.types import StringType, URLType


class License(Model):
    """License information for the exposed API."""

    name = StringType(required=True)  # The license name used for the API.
    url = URLType()  # A URL to the license used for the API. MUST be in the format of a URL.



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from schematics.models import Model

from ._fields import CMDVariantField
from ._selector_index import CMDSelectorIndexField


class CMDSubresourceSelector(Model):
    POLYMORPHIC_KEY = None
    DEFAULT_VARIANT = "$Subresource"

    # properties as tags
    var = CMDVariantField(required=True, default=DEFAULT_VARIANT)
    ref = CMDVariantField(required=True)

    class Options:
        serialize_when_none = False

    @classmethod
    def _claim_polymorphic(cls, data):
        if cls.POLYMORPHIC_KEY is None:
            return False

        if isinstance(data, dict):
            return cls.POLYMORPHIC_KEY in data
        elif isinstance(data, CMDSubresourceSelector):
            return hasattr(data, cls.POLYMORPHIC_KEY)
        return False

    def generate_args(self, ref_args):
        raise NotImplementedError()

    def reformat(self, **kwargs):
        raise NotImplementedError()


class CMDJsonSubresourceSelector(CMDSubresourceSelector):
    """
    : example 1
        <subresourceSelector var="$SubresourceSelector_1" ref="$Instance">
          <json type="object" name="parameters">
            <prop type="array<object>" name="properties.subnets">
              <identifier type="string" name="[Index]" arg="$parameters.properties.subnets[Index]" />
              <item type="object">
                <prop type="array<object>" name="properties.applicationGatewayIpConfigurations">
                  <identifier type="string" name="[].id" arg="$parameters.properties.subnets[].properties.applicationGatewayIpConfigurations[].id"/>
                  <identifier type="string" name="[].name" arg="$parameters.properties.subnets[].properties.applicationGatewayIpConfigurations[].name"/>
                  <item type="object">
                  </item>
                </prop>
              </item>
            </prop>
          </json>
        </subresourceSelector>

    : example 2
        <subresourceSelector var="$SubresourceSelector_2" ref="$Instance">
          <json type="object" name="parameters">
            <prop type="array<object>" name="properties.subnets">
              <identifier type="string" name="[].name" arg="$parameters.properties.subnets[].name" />
              <item type="object">
                <prop type="object" name="properties.applicationGatewayIpConfigurations">
                  <additionalProp>
                    <identifier type="string" name="{Key}" arg="$parameters.properties.subnets[].properties.applicationGatewayIpConfigurations{Key}"/>
                    <item type="object">
                    </item>
                  </additionalProp>
                </prop>
              </item>
            </prop>
          </json>
        </subresourceSelector>

    : example 3
        <subresourceSelector var="$SubresourceSelector_3" ref="$Instance">
          <json type="object" name="parameters">
            <prop type="array<object>" name="properties.subnets">
              <identifier type="string" name="[].name" arg="$parameters.properties.subnets[].name" />
              <item type="object">
                <discriminator property="type" value="AzureFunctionActivity">
                  <prop type="object" name="properties.applicationGatewayIpConfigurations">
                    <additionalProp>
                      <identifier type="string" name="{}.id" arg="$parameters.properties.subnets[].AzureFunctionActivity.properties.applicationGatewayIpConfigurations{}.id"/>
                      <identifier type="string" name="{}.name" arg="$parameters.properties.subnets[].AzureFunctionActivity.properties.applicationGatewayIpConfigurations{}.name"/>
                      <item type="object">
                      </item>
                    </additionalProp>
                  </prop>
                </discriminator>
              </item>
            </prop>
          </json>
        </subresourceSelector>
    """

    POLYMORPHIC_KEY = "json"

    # properties as nodes
    json = CMDSelectorIndexField(required=True)

    def generate_args(self, ref_args):
        return self.json.generate_args(ref_args, "$")

    def reformat(self, **kwargs):
        self.json.reformat(**kwargs)



import click
import logging
from flask import Blueprint
import sys

from swagger.utils.tools import swagger_resource_path_to_resource_id
from utils.config import Config

logger = logging.getLogger('backend')

bp = Blueprint('aaz-cmds', __name__, url_prefix='/AAZ/CMDs', cli_group="command-model")
bp.cli.short_help = "Manage command models in aaz."


def path_type(ctx, param, value):
    import os
    return os.path.expanduser(value)


def resource_id_type(value):
    return swagger_resource_path_to_resource_id(value)


@bp.cli.command("generate-from-swagger", short_help="Generate command models into aaz from swagger specs")
@click.option(
    "--swagger-path", '-s',
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
    default=Config.SWAGGER_PATH,
    callback=Config.validate_and_setup_swagger_path,
    expose_value=False,
    help="The local path of azure-rest-api-specs repo. Official repo is https://github.com/Azure/azure-rest-api-specs"
)
@click.option(
    "--swagger-module-path", "--sm",
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
    default=Config.SWAGGER_MODULE_PATH,
    callback=Config.validate_and_setup_swagger_module_path,
    expose_value=False,
    help="The local path of swagger in module level. It can be substituted for --swagger-path."
)
@click.option(
    "--aaz-path", '-a',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_PATH,
    required=not Config.AAZ_PATH,
    callback=Config.validate_and_setup_aaz_path,
    expose_value=False,
    help="The local path of aaz repo."
)
@click.option(
    "--module", '-m',
    default=Config.DEFAULT_SWAGGER_MODULE,
    required=not Config.DEFAULT_SWAGGER_MODULE,
    callback=Config.validate_and_setup_default_swagger_module,
    expose_value=False,
    help="The name of swagger module."
)
@click.option(
    "--resource-provider", "--rp",
    default=Config.DEFAULT_RESOURCE_PROVIDER,
    required=not Config.DEFAULT_RESOURCE_PROVIDER,
    callback=Config.validate_and_setup_default_resource_provider,
    expose_value=False,
    help="The resource provider name."
)
@click.option(
    "--swagger-tag", "--tag",
    required=True,
    help="Swagger tag with input files."
)
@click.option(
    "--workspace-path",
    help="The path to export the workspace for modification."
)
def generate_command_models_from_swagger(swagger_tag, workspace_path=None):
    from swagger.controller.specs_manager import SwaggerSpecsManager
    from command.controller.specs_manager import AAZSpecsManager
    from command.controller.workspace_manager import WorkspaceManager
    from utils.config import Config
    from utils.exceptions import InvalidAPIUsage
    from command.model.configuration import CMDHelp

    try:
        swagger_specs = SwaggerSpecsManager()
        aaz_specs = AAZSpecsManager()

        module_manager = swagger_specs.get_module_manager(Config.DEFAULT_PLANE, Config.DEFAULT_SWAGGER_MODULE)
        rp = module_manager.get_resource_provider(Config.DEFAULT_RESOURCE_PROVIDER)

        resource_map = rp.get_resource_map_by_tag(swagger_tag)
        if not resource_map:
            raise InvalidAPIUsage(f"Tag `{swagger_tag}` is not exist")

        version_resource_map = {}
        for resource_id, version_map in resource_map.items():
            v_list = [v for v in version_map]
            if len(v_list) > 1:
                raise InvalidAPIUsage(f"Tag `{swagger_tag}` contains multiple api versions of one resource", payload={
                    "Resource": resource_id,
                    "versions": v_list,
                })
            v = v_list[0]
            if v not in version_resource_map:
                version_resource_map[v] = []
            version_resource_map[v].append({
                "id": resource_id
            })

        ws = WorkspaceManager.new(
            name=Config.DEFAULT_SWAGGER_MODULE,
            plane=Config.DEFAULT_PLANE,
            folder=workspace_path or WorkspaceManager.IN_MEMORY,  # if workspace path exist, use workspace else use in memory folder
            swagger_manager=swagger_specs,
            aaz_manager=aaz_specs,
        )
        mod_names = Config.DEFAULT_SWAGGER_MODULE.split('/')
        for version, resources in version_resource_map.items():
            ws.add_new_resources_by_swagger(
                mod_names=mod_names, version=version, resources=resources
            )

        # provide default short summary
        for node in ws.iter_command_tree_nodes():
            if not node.help:
                node.help = CMDHelp()
            if not node.help.short:
                node.help.short = f"Manage {node.names[-1]}"

        for leaf in ws.iter_command_tree_leaves():
            if not leaf.help:
                leaf.help = CMDHelp()
            if not leaf.help.short:
                n = leaf.names[-1]
                n = n[0].upper() + n[1:]
                leaf.help.short = f"{n} {leaf.names[-2]}"

        if not ws.is_in_memory:
            ws.save()

        ws.generate_to_aaz()

    except InvalidAPIUsage as err:
        logger.error(err)
        sys.exit(1)
    except ValueError as err:
        logger.error(err)
        sys.exit(1)



from ._arg import CMDArgEnumItem, CMDArgEnum, \
    CMDArgDefault, CMDArgBlank, \
    CMDArgBase, CMDArg, \
    CMDClsArgBase, CMDClsArg, \
    CMDStringArgBase, CMDStringArg, \
    CMDByteArgBase, CMDByteArg, \
    CMDBinaryArgBase, CMDBinaryArg, \
    CMDDurationArgBase, CMDDurationArg, \
    CMDDateArgBase, CMDDateArg, \
    CMDDateTimeArgBase, CMDDateTimeArg, \
    CMDUuidArgBase, CMDUuidArg, \
    CMDPasswordArgBase, CMDPasswordArg, \
    CMDSubscriptionIdArgBase, CMDSubscriptionIdArg, \
    CMDResourceGroupNameArgBase, CMDResourceGroupNameArg, \
    CMDResourceIdArgBase, CMDResourceIdArg, \
    CMDResourceLocationArgBase, CMDResourceLocationArg, \
    CMDIntegerArgBase, CMDIntegerArg, \
    CMDInteger32ArgBase, CMDInteger32Arg, CMDInteger64ArgBase, CMDInteger64Arg, \
    CMDBooleanArgBase, CMDBooleanArg, \
    CMDFloatArgBase, CMDFloatArg, \
    CMDFloat32ArgBase, CMDFloat32Arg, CMDFloat64ArgBase, CMDFloat64Arg, \
    CMDObjectArgBase, CMDObjectArg, \
    CMDArrayArgBase, CMDArrayArg, \
    CMDArgPromptInput, CMDPasswordArgPromptInput
from ._arg_builder import CMDArgBuilder
from ._arg_group import CMDArgGroup
from ._command import CMDCommand
from ._command_group import CMDCommandGroup
from ._condition import CMDConditionOperator, \
    CMDConditionAndOperator, CMDConditionOrOperator, CMDConditionNotOperator, CMDConditionHasValueOperator, \
    CMDCondition
from ._configuration import CMDConfiguration
from ._content import CMDRequestJson, CMDResponseJson
from ._example import CMDCommandExample
from ._fields import CMDBooleanField, CMDStageField, CMDVariantField, CMDClassField, \
    CMDPrimitiveField, CMDRegularExpressionField, CMDVersionField, CMDResourceIdField, CMDCommandNameField, \
    CMDCommandGroupNameField, CMDURLPathField, CMDConfirmation
from ._format import CMDStringFormat, CMDIntegerFormat, CMDFloatFormat, CMDObjectFormat, CMDArrayFormat, \
    CMDResourceIdFormat
from ._help import CMDHelp, CMDArgumentHelp
from ._http import CMDHttpRequestArgs, CMDHttpRequestPath, CMDHttpRequestQuery, CMDHttpRequestHeader, \
    CMDHttpRequest, \
    CMDHttpResponseHeaderItem, CMDHttpResponseHeader, CMDHttpResponse, \
    CMDHttpAction
from ._http_request_body import CMDHttpRequestBody, CMDHttpRequestJsonBody
from ._http_response_body import CMDHttpResponseBody, CMDHttpResponseJsonBody
from ._instance_create import CMDInstanceCreateAction, CMDJsonInstanceCreateAction
from ._instance_delete import CMDInstanceDeleteAction, CMDJsonInstanceDeleteAction
from ._instance_update import CMDInstanceUpdateAction, CMDJsonInstanceUpdateAction
from ._operation import CMDOperation, CMDHttpOperation, CMDInstanceUpdateOperation, CMDInstanceCreateOperation, \
    CMDInstanceDeleteOperation, CMDHttpOperationLongRunning
from ._output import CMDOutput, CMDObjectOutput, CMDArrayOutput, CMDStringOutput
from ._resource import CMDResource
from ._schema import CMDSchemaEnumItem, CMDSchemaEnum, CMDSchemaDefault, \
    CMDSchemaBase, CMDSchema, \
    CMDClsSchemaBase, CMDClsSchema, \
    CMDStringSchemaBase, CMDStringSchema, \
    CMDByteSchemaBase, CMDByteSchema, \
    CMDBinarySchemaBase, CMDBinarySchema, \
    CMDDurationSchemaBase, CMDDurationSchema, \
    CMDDateSchemaBase, CMDDateSchema, \
    CMDDateTimeSchemaBase, CMDDateTimeSchema, \
    CMDUuidSchemaBase, CMDUuidSchema, \
    CMDPasswordSchemaBase, CMDPasswordSchema, \
    CMDResourceIdSchemaBase, CMDResourceIdSchema, \
    CMDResourceLocationSchemaBase, CMDResourceLocationSchema, \
    CMDIntegerSchemaBase, CMDIntegerSchema, \
    CMDInteger32SchemaBase, CMDInteger32Schema, \
    CMDInteger64SchemaBase, CMDInteger64Schema, \
    CMDBooleanSchemaBase, CMDBooleanSchema, \
    CMDFloatSchemaBase, CMDFloatSchema, \
    CMDFloat32SchemaBase, CMDFloat32Schema, \
    CMDFloat64SchemaBase, CMDFloat64Schema, \
    CMDObjectSchemaDiscriminator, CMDObjectSchemaAdditionalProperties, CMDObjectSchemaBase, CMDObjectSchema, \
    CMDIdentityObjectSchemaBase, CMDIdentityObjectSchema, \
    CMDArraySchemaBase, CMDArraySchema
from ._selector_index import CMDSelectorIndexBase, CMDSelectorIndex, CMDObjectIndexDiscriminator, \
    CMDObjectIndexAdditionalProperties, CMDObjectIndexBase, CMDObjectIndex, CMDArrayIndexBase, CMDArrayIndex
from ._subresource_selector import CMDSubresourceSelector, CMDJsonSubresourceSelector
from ._utils import CMDDiffLevelEnum, DEFAULT_CONFIRMATION_PROMPT
from ._xml import XMLSerializer



from command.model.configuration import CMDCommand
from schematics.types import StringType, ModelType, BooleanType
from utils.config import Config


class CLIProfileNameField(StringType):
    """The stage for command group, command or argument."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            choices=tuple(Config.CLI_PROFILES),
            default=Config.CLI_DEFAULT_PROFILE,
            serialize_when_none=False,
            *args,
            **kwargs
        )


class CLICommandNameField(StringType):

    def __init__(self, *args, **kwargs):
        super().__init__(
            regex=r'^[a-z0-9]+(-[a-z0-9]+)*$',
            min_length=1, *args, **kwargs)


class CLICommandConfigurationField(ModelType):

    def __init__(self, *args, **kwargs):
        super().__init__(
            CMDCommand,
            serialize_when_none=False,
        )

    def to_primitive(self, value, context=None):
        return None  # return None when value is false to hide field with `serialize_when_none=False`


class CLIVersionField(StringType):

    def __init__(self, *args, **kwargs):
        super(CLIVersionField, self).__init__(*args, **kwargs)


class CLIResourceIdField(StringType):

    def __init__(self, *args, **kwargs):
        super(CLIResourceIdField, self).__init__(
            serialized_name='id',
            deserialize_from='id',
            *args, **kwargs
        )


class CLIModifiedField(BooleanType):

    def to_primitive(self, value, context=None):
        # ignore modified value when using to primitive
        return None



from schematics.models import Model
from schematics.types import StringType, ModelType
from .contact import Contact
from .license import License
from .fields import XmsCodeGenerationSettingsField, XcadlGeneratedField, XTypespecGeneratedField


class Info(Model):
    """The object provides metadata about the API. The metadata can be used by the clients if needed, and can be presented in the Swagger-UI for convenience."""

    title = StringType(required=True)       # The title of the application.
    version = StringType(required=True)     # Provides the version of the application API (not to be confused with the specification version)
    description = StringType()  # A short description of the application
    terms_of_service = StringType(
        serialized_name="termsOfService",
        deserialize_from="termsOfService"
    )  # The Terms of Service for the API.
    contact = ModelType(Contact)  # The contact information for the exposed API.
    license = ModelType(License)  # The license information for the exposed API.

    # specific properties
    _x_ms_code_generation_settings = XmsCodeGenerationSettingsField()  # Deprecated
    _x_cadl_generated = XcadlGeneratedField()
    _x_typespec_generated = XTypespecGeneratedField()



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

URL_PARAMETER_PLACEHOLDER = "{}"


def swagger_resource_path_to_resource_id_template(path):
    path_parts = path.split("?", maxsplit=1)
    url_parts = path_parts[0].split("/")
    idx = 2  # ignore to the parameter name in first part of url, such as `/{parameter}` or `/{parameter}/...`
    while idx < len(url_parts):
        if url_parts[idx].startswith("{") and url_parts[idx].endswith("}"):
            url_parts[idx] = URL_PARAMETER_PLACEHOLDER
        idx += 1
    path_parts[0] = "/".join(url_parts)
    return "?".join(path_parts)


def swagger_resource_path_to_resource_id(path):
    path_parts = path.split("?", maxsplit=1)
    url_parts = path_parts[0].split("/")
    idx = 2  # ignore to the parameter name in first part of url, such as `/{parameter}` or `/{parameter}/...`
    while idx < len(url_parts):
        if url_parts[idx].startswith("{") and url_parts[idx].endswith("}"):
            url_parts[idx] = URL_PARAMETER_PLACEHOLDER
        idx += 1
    path_parts[0] = "/".join(url_parts).lower()
    return "?".join(path_parts)



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

import sys
from os.path import abspath, dirname, join

# mark sources root
sys.path.insert(0, join(abspath(dirname(__file__))))



from schematics.models import Model
from schematics.types import StringType, ModelType
from .external_documentation import ExternalDocumentation


class Tag(Model):
    """Allows adding meta data to a single tag that is used by the Operation Object. It is not mandatory to have a Tag Object per tag used there."""

    name = StringType(required=True)  # The name of the tag.
    description = StringType()  # A short description for the tag. GFM syntax can be used for rich text representation.
    external_docs = ModelType(
        ExternalDocumentation,
        serialized_name="externalDocs",
        deserialize_from="externalDocs"
    )  # Additional external documentation for this tag.



from schematics.models import Model
from schematics.types import StringType

from ._fields import CMDResourceIdField, CMDVersionField
from utils.base64 import b64decode_str


class CMDResource(Model):
    # properties as tags
    id = CMDResourceIdField(required=True)
    version = CMDVersionField(required=True)
    subresource = StringType()  # subresource index, used for sub commands generation.

    swagger = StringType(required=True)  # swagger resource provider, <plane>/<path:mod_names>/ResourceProviders/<rp_name>/Paths/<base64Encoded Path>/V/<base64Encoded version>

    class Options:
        serialize_when_none = False

    @property
    def plane(self):
        return self.swagger.split('/')[0]

    @property
    def mod_names(self):
        return self.swagger.split("/ResourceProviders/")[0].split('/')[1:]

    @property
    def rp_name(self):
        return self.swagger.split("/ResourceProviders/")[1].split('/')[0]

    @property
    def swagger_path(self):
        return b64decode_str(self.swagger.split("/Paths/")[1].split('/')[0])



from cli.model.atomic import CLIAtomicCommand
from command.model.configuration import CMDCommand, CMDHttpOperation, CMDCondition, CMDConditionAndOperator, \
    CMDConditionOrOperator, CMDConditionNotOperator, CMDConditionHasValueOperator, CMDInstanceUpdateOperation, \
    CMDJsonInstanceUpdateAction, CMDResourceGroupNameArg, CMDJsonSubresourceSelector, CMDInstanceCreateOperation, \
    CMDInstanceDeleteOperation, CMDJsonInstanceCreateAction, CMDJsonInstanceDeleteAction
from utils.case import to_camel_case, to_snack_case
from utils.plane import PlaneEnum
from .az_operation_generator import AzHttpOperationGenerator, AzJsonUpdateOperationGenerator, \
    AzGenericUpdateOperationGenerator, AzRequestClsGenerator, AzResponseClsGenerator, \
    AzInstanceUpdateOperationGenerator, AzLifeCycleInstanceUpdateCallbackGenerator, AzJsonCreateOperationGenerator, \
    AzJsonDeleteOperationGenerator, AzLifeCycleCallbackGenerator
from .az_arg_group_generator import AzArgGroupGenerator, AzArgClsGenerator
from .az_output_generator import AzOutputGenerator
from .az_selector_generator import AzJsonSelectorGenerator
from utils import exceptions
import logging


logger = logging.getLogger('backend')


class AzCommandCtx:

    def __init__(self):
        self._cls_arg_maps = {}
        self._ctx_arg_map = {}
        self._selectors = {}
        self.rg_arg_var = None

        self.arg_clses = {}
        self.update_clses = {}
        self.response_clses = {}
        self.support_id_part = True

    def set_argument_cls(self, arg):
        cls_name = arg.cls
        self._cls_arg_maps[f"@{cls_name}"] = {}
        assert cls_name not in self.arg_clses, f"Argument class {cls_name} is defined more than one place"
        self.arg_clses[cls_name] = AzArgClsGenerator(cls_name, self, arg)

    def set_argument(self, keys, arg, ctx_namespace='self.ctx.args'):
        var_name = arg.var
        hide = arg.hide
        if var_name.startswith('@'):
            map_name = var_name.replace('[', '.[').replace('{', '.{').split('.', maxsplit=1)[0]
            if map_name != keys[0]:
                raise exceptions.VerificationError(
                    "Invalid argument var",
                    details=f"argument var '{var_name}' does not start with '{keys[0]}'"
                )
            if map_name not in self._cls_arg_maps:
                self._cls_arg_maps[map_name] = {}
            self._cls_arg_maps[map_name][var_name] = (
                '.'.join(keys).replace('.[', '[').replace('.{', '{'),
                hide
            )
        else:
            self._ctx_arg_map[var_name] = (
                '.'.join([ctx_namespace, *keys]).replace('.[', '[').replace('.{', '{'),
                hide
            )

        if isinstance(arg, CMDResourceGroupNameArg):
            assert self.rg_arg_var is None, "Resource Group Argument defined twice"
            self.rg_arg_var = arg.var

    def get_argument(self, var_name):
        if var_name.startswith('@'):
            map_name = var_name.replace('[', '.[').replace('{', '.{').split('.', maxsplit=1)[0]
            if map_name not in self._cls_arg_maps:
                raise exceptions.VerificationError(
                    "Invalid argument var",
                    details=f"argument var '{var_name}' has unregistered class '{map_name}'."
                )
            if var_name not in self._cls_arg_maps[map_name]:
                raise exceptions.VerificationError(
                    "Invalid argument var",
                    details=f"argument var '{var_name}' does not find."
                )
            return self._cls_arg_maps[map_name][var_name]
        else:
            if var_name not in self._ctx_arg_map:
                raise exceptions.VerificationError(
                    "Invalid argument var",
                    details=f"argument var '{var_name}' does not find."
                )
            return self._ctx_arg_map[var_name]

    def get_variant(self, variant, name_only=False):
        if variant.startswith('$'):
            variant = variant[1:]
        variant = to_snack_case(variant)
        if name_only:
            return variant

        is_selector = variant in self._selectors

        if is_selector:
            return f'self.ctx.selectors.{variant}', is_selector
        else:
            return f'self.ctx.vars.{variant}', is_selector

    def set_update_cls(self, schema):
        cls_name = schema.cls
        assert cls_name not in self.update_clses, f"Schema cls '{cls_name}', is defined more than once"
        self.update_clses[cls_name] = AzRequestClsGenerator(self, cls_name, schema)

    def set_response_cls(self, schema):
        cls_name = schema.cls
        assert cls_name not in self.response_clses, f"Schema cls '{cls_name}', is defined more than once"
        self.response_clses[cls_name] = AzResponseClsGenerator(self, cls_name, schema)

    def set_selector(self, selector):
        self._selectors[self.get_variant(selector.var, name_only=True)] = selector

    def render_arg_resource_id_template(self, template):
        # TODO: fill blank placeholders as much as possible

        return template


class AzCommandGenerator:

    ARGS_SCHEMA_NAME = "_args_schema"

    def __init__(self, cmd: CLIAtomicCommand, is_wait=False):
        self.cmd = cmd
        self.is_wait = is_wait
        self.cmd_ctx = AzCommandCtx()

        if cmd.names[-1] in ("create", "list"):
            # disable id part for create and list command
            self.cmd_ctx.support_id_part = False

        assert isinstance(self.cmd.cfg, CMDCommand)
        self.conditions = []
        if self.cmd.cfg.conditions:
            for idx, condition in enumerate(self.cmd.cfg.conditions):
                self.conditions.append((condition.var, f"condition_{idx}", condition))

        # prepare arguments
        self.arg_groups = []
        if self.cmd.cfg.arg_groups:
            for arg_group in self.cmd.cfg.arg_groups:
                if arg_group.args:
                    self.arg_groups.append(AzArgGroupGenerator(self.ARGS_SCHEMA_NAME, self.cmd_ctx, arg_group))

        self.selectors = []
        if self.cmd.cfg.subresource_selector:
            # disable id part for subresource command
            self.cmd_ctx.support_id_part = False

            if isinstance(self.cmd.cfg.subresource_selector, CMDJsonSubresourceSelector):
                selector = AzJsonSelectorGenerator(self.cmd_ctx, self.cmd.cfg.subresource_selector)
                self.selectors.append(selector)
            else:
                raise NotImplementedError()

        # prepare operations
        self.lro_counts = 0
        self.operations = []
        self.http_operations = []
        self.json_instance_operations = []
        self.support_generic_update = False

        json_instance_counts = 0
        for operation in self.cmd.cfg.operations:
            lr = False
            if isinstance(operation, CMDHttpOperation):
                op_cls_name = to_camel_case(operation.operation_id)
                if operation.long_running:
                    lr = True
                op = AzHttpOperationGenerator(op_cls_name, self.cmd_ctx, operation)
                self.http_operations.append(op)
            elif isinstance(operation, CMDInstanceUpdateOperation):
                if isinstance(operation.instance_update, CMDJsonInstanceUpdateAction):
                    op_cls_name = f'InstanceUpdateByJson'
                    if json_instance_counts > 0:
                        op_cls_name += f'_{json_instance_counts}'
                    op = AzJsonUpdateOperationGenerator(op_cls_name, self.cmd_ctx, operation)
                    self.json_instance_operations.append(op)
                    json_instance_counts += 1
                else:
                    raise NotImplementedError()
            elif isinstance(operation, CMDInstanceCreateOperation):
                if isinstance(operation.instance_create, CMDJsonInstanceCreateAction):
                    op_cls_name = f"InstanceCreateByJson"
                    if json_instance_counts > 0:
                        op_cls_name += f"_{json_instance_counts}"
                    op = AzJsonCreateOperationGenerator(op_cls_name, self.cmd_ctx, operation)
                    self.json_instance_operations.append(op)
                    json_instance_counts += 1
                else:
                    raise NotImplementedError()
            elif isinstance(operation, CMDInstanceDeleteOperation):
                if isinstance(operation.instance_delete, CMDJsonInstanceDeleteAction):
                    op_cls_name = f'InstanceDeleteByJson'
                    if json_instance_counts > 0:
                        op_cls_name += f'_{json_instance_counts}'
                    op = AzJsonDeleteOperationGenerator(op_cls_name, self.cmd_ctx, operation)
                    self.json_instance_operations.append(op)
                    json_instance_counts += 1
            else:
                raise NotImplementedError()

            if lr:
                self.lro_counts += 1
            self.operations.append(op)

        # generic_update_op
        if self.json_instance_operations:
            self.support_generic_update = self.cmd.names[-1] == "update"
            if self.support_generic_update:
                # make sure all json update operations has the same variant key
                variant_key = self.json_instance_operations[0].variant_key
                is_selector_variant = self.json_instance_operations[0].is_selector_variant
                for op in self.json_instance_operations[1:]:
                    if not isinstance(op, AzJsonUpdateOperationGenerator) or op.variant_key != variant_key:
                        self.support_generic_update = False
                        break
            if self.support_generic_update:
                min_idx = None
                max_idx = None
                for idx, op in enumerate(self.operations):
                    if isinstance(op, AzJsonUpdateOperationGenerator):
                        if min_idx is None:
                            min_idx = idx
                        if max_idx is None or max_idx < idx:
                            max_idx = idx
                if max_idx + 1 - min_idx != len(self.json_instance_operations):
                    # has other operations between
                    self.support_generic_update = False
                else:
                    op = AzGenericUpdateOperationGenerator(self.cmd_ctx, variant_key, is_selector_variant)
                    self.json_instance_operations.append(op)
                    self.operations = [*self.operations[:max_idx+1], op, *self.operations[max_idx+1:]]

        # Add instance update callbacks
        first_instance_op_idx = None
        last_instance_op_idx = None
        for idx, op in enumerate(self.operations):
            if isinstance(op, AzInstanceUpdateOperationGenerator):
                if first_instance_op_idx is None:
                    first_instance_op_idx = idx
                last_instance_op_idx = idx
        if last_instance_op_idx is not None and len(self.operations) > last_instance_op_idx + 1:
            post_op_generator = AzLifeCycleInstanceUpdateCallbackGenerator('post_instance_update', self.operations[last_instance_op_idx].variant_key, self.operations[last_instance_op_idx].is_selector_variant)
            self.operations = [*self.operations[:last_instance_op_idx+1], post_op_generator, *self.operations[last_instance_op_idx+1:]]
        if first_instance_op_idx is not None and first_instance_op_idx > 0:
            pre_op_generator = AzLifeCycleInstanceUpdateCallbackGenerator('pre_instance_update', self.operations[first_instance_op_idx].variant_key, self.operations[last_instance_op_idx].is_selector_variant)
            self.operations = [*self.operations[:first_instance_op_idx], pre_op_generator, *self.operations[first_instance_op_idx:]]

        # Add instance create callbacks
        first_instance_op_idx = None
        last_instance_op_idx = None
        for idx, op in enumerate(self.operations):
            if isinstance(op, AzJsonCreateOperationGenerator):
                if first_instance_op_idx is None:
                    first_instance_op_idx = idx
                last_instance_op_idx = idx
        if last_instance_op_idx is not None and len(self.operations) > last_instance_op_idx + 1:
            post_op_generator = AzLifeCycleInstanceUpdateCallbackGenerator('post_instance_create', self.operations[last_instance_op_idx].variant_key, self.operations[last_instance_op_idx].is_selector_variant)
            self.operations = [*self.operations[:last_instance_op_idx+1], post_op_generator, *self.operations[last_instance_op_idx+1:]]
        if first_instance_op_idx is not None and first_instance_op_idx > 0:
            pre_op_generator = AzLifeCycleCallbackGenerator('pre_instance_create')
            self.operations = [*self.operations[:first_instance_op_idx], pre_op_generator, *self.operations[first_instance_op_idx:]]

        # Add instance delete callbacks
        first_instance_op_idx = None
        last_instance_op_idx = None
        for idx, op in enumerate(self.operations):
            if isinstance(op, AzJsonDeleteOperationGenerator):
                if first_instance_op_idx is None:
                    first_instance_op_idx = idx
                last_instance_op_idx = idx
        if last_instance_op_idx is not None and len(self.operations) > last_instance_op_idx + 1:
            post_op_generator = AzLifeCycleCallbackGenerator('post_instance_delete')
            self.operations = [*self.operations[:last_instance_op_idx + 1], post_op_generator,
                               *self.operations[last_instance_op_idx + 1:]]
        if first_instance_op_idx is not None and first_instance_op_idx > 0:
            pre_op_generator = AzLifeCycleCallbackGenerator('pre_instance_delete')
            self.operations = [*self.operations[:first_instance_op_idx], pre_op_generator,
                               *self.operations[first_instance_op_idx:]]

        self.plane = None
        for resource in self.cmd.resources:
            if not self.plane:
                self.plane = resource.plane
            elif resource.plane != self.plane:
                raise ValueError(f"Find multiple planes in a command: {resource.plane}, {self.plane}")

        self.client_type = PlaneEnum.http_client(self.plane)

        # prepare outputs
        self.outputs = []
        self.paging = False
        if self.cmd.cfg.outputs:
            for output in self.cmd.cfg.outputs:
                output_generator = AzOutputGenerator(output, self.cmd_ctx)
                self.outputs.append(output_generator)
                if output_generator.next_link:
                    self.paging = True
        if self.lro_counts > 0 and self.paging:
            self.paging = False
            # TODO: support paging for long running command later
            logger.warning(f"Disable paging for long running command: '{self.name}'")

        if len(self.outputs) > 1:
            # TODO: add support for output larger than 1
            raise NotImplementedError()

        if self.paging:
            # disable id part for paging commands
            self.cmd_ctx.support_id_part = False

    @property
    def name(self):
        return ' '.join(self.cmd.names)

    @property
    def cls_name(self):
        return to_camel_case(self.cmd.names[-1])

    @property
    def helper_cls_name(self):
        return f'_{to_camel_case(self.cmd.names[-1])}Helper'

    @property
    def help(self):
        return self.cmd.help

    @property
    def register_info(self):
        return self.cmd.register_info

    @property
    def support_no_wait(self):
        return self.cmd.support_no_wait or False

    @property
    def version(self):
        return self.cmd.version

    @property
    def resources(self):
        return self.cmd.resources

    def get_arg_clses(self):
        return sorted(self.cmd_ctx.arg_clses.values(), key=lambda a: a.name)

    def get_update_clses(self):
        return sorted(self.cmd_ctx.update_clses.values(), key=lambda s: s.name)

    def get_response_clses(self):
        return sorted(self.cmd_ctx.response_clses.values(), key=lambda s: s.name)

    def has_outputs(self):
        return len(self.outputs) > 0

    def render_condition(self, condition):
        assert isinstance(condition, CMDCondition)
        return self._render_operator(condition.operator, parent_priority=0)

    def _render_operator(self, operator, parent_priority):
        if isinstance(operator, CMDConditionOrOperator):
            results = []
            for op in operator.operators:
                results.append(
                    self._render_operator(op, parent_priority=2 if len(operator.operators) > 1 else parent_priority)
                )
            if len(results) < 1:
                raise ValueError()
            elif len(results) == 1:
                result = results[0]
            else:
                result = ' or '.join(results)
                if parent_priority > 2:
                    result = f"({result})"
        elif isinstance(operator, CMDConditionAndOperator):
            results = []
            for op in operator.operators:
                results.append(
                    self._render_operator(op, parent_priority=3 if len(operator.operators) > 1 else parent_priority)
                )
            if len(results) < 1:
                raise ValueError()
            elif len(results) == 1:
                result = results[0]
            else:
                result = ' and '.join(results)
                if parent_priority > 3:
                    result = f"({result})"
        elif isinstance(operator, CMDConditionNotOperator):
            result = f'{self._render_operator(operator.operator, parent_priority=6)} is not True'
            if parent_priority > 6:
                result = f"({result})"
        elif isinstance(operator, CMDConditionHasValueOperator):
            arg_keys, hide = self.cmd_ctx.get_argument(operator.arg)
            assert not hide
            result = f'has_value({arg_keys})'
        else:
            raise NotImplementedError()
        return result

    def render_operation_when(self, when):
        results = []
        for condition_var in when:
            condition_name = None
            for c_v, c_name, _ in self.conditions:
                if c_v == condition_var:
                    condition_name = c_name
                    break
            if not condition_name:
                raise KeyError(f"Condition variant not exist: {condition_var}")
            results.append(condition_name)
        return ' or '.join(results)



from collections import OrderedDict

from swagger.model.specs import SwaggerSpecs, SingleModuleSwaggerSpecs, ResourceProvider, SwaggerModule
from utils import exceptions
from utils.config import Config
from utils.plane import PlaneEnum


class SwaggerSpecsModuleManager:

    def __init__(self, plane, module):
        self.plane = plane
        self.module = module
        self._rps_catch = None
        self._resource_op_group_map_cache = {}
        self._resource_map_cache = {}
        assert plane in PlaneEnum.choices(), f"Invalid plane: '{self.plane}'"
        assert isinstance(module, SwaggerModule), f"Invalid module type: '{type(module)}'"

    def get_resource_providers(self):
        if self._rps_catch is None:
            self._rps_catch = self.module.get_resource_providers()
        return self._rps_catch

    def get_resource_provider(self, rp_name):
        rps = self.get_resource_providers()
        rp = None
        for v in rps:
            if v.name == rp_name:
                rp = v
                break
        if rp is None:

            raise exceptions.ResourceNotFind(f"resource provider not find '{rp_name}'")
        return rp

    def get_grouped_resource_map(self, rp_name):
        key = rp_name
        if key in self._resource_op_group_map_cache:
            return self._resource_op_group_map_cache[key]

        rp = self.get_resource_provider(rp_name)
        resource_map = self.get_resource_map(rp)
        resource_op_group_map = OrderedDict()
        for resource_id, version_map in resource_map.items():
            op_group_name = self.get_resource_op_group_name(version_map)
            if op_group_name not in resource_op_group_map:
                resource_op_group_map[op_group_name] = OrderedDict()
            resource_op_group_map[op_group_name][resource_id] = version_map
        self._resource_op_group_map_cache[key] = resource_op_group_map
        return self._resource_op_group_map_cache[key]

    @staticmethod
    def get_resource_op_group_name(version_map):
        _, latest_resource = sorted(
            version_map.items(),
            key=lambda item: str(item[0]),
            reverse=True
        )[0]
        return latest_resource.get_operation_group_name() or ""

    def get_resource_version_map(self, resource_id, rp_name=None):
        if rp_name:
            rps = [self.get_resource_provider(rp_name)]
        else:
            rps = self.get_resource_providers()

        version_maps = []
        resource_rps = []
        for rp in rps:
            resource_map = self.get_resource_map(rp)
            if resource_id in resource_map:
                version_maps.append(resource_map[resource_id])
                resource_rps.append(rp)

        if not version_maps:
            raise exceptions.ResourceNotFind(f"Resource not find '{resource_id}'")

        if len(resource_rps) > 1:
            raise exceptions.InvalidAPIUsage(
                "Find multiple resource providers: \n" +
                "\n".join([r.folder_path for r in resource_rps])
            )
        return version_maps[0]

    def get_resource_in_version(self, resource_id, version, rp_name=None):
        if rp_name:
            rps = [self.get_resource_provider(rp_name)]
        else:
            rps = self.get_resource_providers()

        resources = []
        for rp in rps:
            resource_map = self.get_resource_map(rp)
            if resource_id in resource_map and version in resource_map[resource_id]:
                resources.append(resource_map[resource_id][version])

        if not resources:
            raise exceptions.ResourceNotFind(f"Resource not find '{resource_id}' '{version}'")

        if len(resources) > 1:
            raise exceptions.InvalidAPIUsage(
                "Find multiple resources in files: \n" +
                "\n".join([r.file_path for r in resources])
            )

        return resources[0]

    def get_resource_map(self, rp):
        assert isinstance(rp, ResourceProvider)
        key = str(rp)
        if key not in self._resource_map_cache:
            self._resource_map_cache[key] = rp.get_resource_map()
        return self._resource_map_cache[key]


class SwaggerSpecsManager:

    def __init__(self):
        if Config.SWAGGER_PATH:
            self.specs = SwaggerSpecs(folder_path=Config.SWAGGER_PATH)
        elif Config.SWAGGER_MODULE_PATH:
            if not Config.DEFAULT_SWAGGER_MODULE:
                raise ValueError("SWAGGER_MODULE is required when using SWAGGER_MODULE_PATH")
            self.specs = SingleModuleSwaggerSpecs(
                folder_path=Config.SWAGGER_MODULE_PATH, module_name=Config.DEFAULT_SWAGGER_MODULE)
        else:
            raise ValueError("Require SWAGGER_PATH or SWAGGER_MODULE_PATH")

        self._modules_cache = {}
        self._module_managers_cache = {}

    def get_modules(self, plane):
        if plane in self._modules_cache:
            return self._modules_cache[plane]

        if plane == PlaneEnum.Mgmt:
            modules = self.specs.get_mgmt_plane_modules(plane=plane)
        elif plane in PlaneEnum.choices():
            modules = self.specs.get_data_plane_modules(plane=plane)
        else:
            raise exceptions.InvalidAPIUsage(f"invalid plane name '{plane}'")

        result = OrderedDict()
        for m in modules:
            for rp in m.get_resource_providers():
                module = rp.swagger_module
                module_str = str(module)
                if module_str not in result:
                    result[module_str] = module

        self._modules_cache[plane] = [*result.values()]
        return self._modules_cache[plane]

    def get_module(self, plane, mod_names):
        if isinstance(mod_names, str):
            mod_names = mod_names.split('/')

        if plane == PlaneEnum.Mgmt:
            module = self.specs.get_mgmt_plane_module(*mod_names, plane=plane)
        elif plane in PlaneEnum.choices():
            module = self.specs.get_data_plane_module(*mod_names, plane=plane)
        else:
            raise exceptions.InvalidAPIUsage(f"invalid plane name '{plane}'")

        if not module:
            raise exceptions.ResourceNotFind(f"Module not find '{'/'.join(mod_names)}'")
        return module

    def get_module_manager(self, plane, mod_names, without_catch=False) -> SwaggerSpecsModuleManager:
        key = (plane, tuple(mod_names))
        if without_catch or key not in self._module_managers_cache:
            module = self.get_module(plane, mod_names)
            self._module_managers_cache[key] = SwaggerSpecsModuleManager(plane, module)

        return self._module_managers_cache[key]



import os

from utils.plane import PlaneEnum
from ._resource_provider import ResourceProvider


class SwaggerModule:

    def __init__(self, plane, name, folder_path, parent=None):
        assert plane in PlaneEnum.choices()
        self.plane = plane
        self.name = name
        self.folder_path = folder_path
        self._parent = parent

    def __str__(self):
        if self._parent is not None:
            return f"{self._parent}/{self.name}"
        else:
            return f"{self.name}"

    def __eq__(self, other):
        return isinstance(other, self.__class__) and str(other) == str(self)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(str(self))

    @property
    def names(self):
        if self._parent is None:
            return [self.name]
        else:
            return [*self._parent.names, self.name]


class MgmtPlaneModule(SwaggerModule):

    def __str__(self):
        if self._parent is None:
            return f'{self.plane}/{super().__str__()}'
        else:
            return super(MgmtPlaneModule, self).__str__()

    def get_resource_providers(self):
        rp = []
        for name in os.listdir(self.folder_path):
            path = os.path.join(self.folder_path, name)
            if os.path.isdir(path):
                name_parts = name.split('.')
                if len(name_parts) >= 2:
                    readme_path = _search_readme_md_path(path, search_parent=True)
                    rp.append(ResourceProvider(name, path, readme_path, swagger_module=self))
                elif name.lower() != 'common':
                    # azsadmin module only
                    sub_module = MgmtPlaneModule(plane=self.plane, name=name, folder_path=path, parent=self)
                    rp.extend(sub_module.get_resource_providers())
        return rp


class DataPlaneModule(SwaggerModule):

    def __str__(self):
        if self._parent is None:
            return f'{self.plane}/{super().__str__()}'
        else:
            return super(DataPlaneModule, self).__str__()

    def get_resource_providers(self):
        rp = []
        for name in os.listdir(self.folder_path):
            path = os.path.join(self.folder_path, name)
            if os.path.isdir(path):
                if name.lower() in ('preview', 'stable'):
                    readme_path = _search_readme_md_path(self.folder_path)
                    rp = [ResourceProvider(self.name, self.folder_path, readme_path, swagger_module=self)]
                    break
                name_parts = name.split('.')
                if len(name_parts) >= 2:
                    readme_path = _search_readme_md_path(path, search_parent=True)
                    rp.append(ResourceProvider(name, path, readme_path, swagger_module=self))
                elif name.lower() != 'common':
                    has_sub_module = True
                    for sub_name in os.listdir(path):
                        sub_module_path = os.path.join(path, sub_name)
                        if os.path.isdir(sub_module_path) and sub_name.lower() in ('preview', 'stable'):
                            has_sub_module = False
                            break
                    if has_sub_module:
                        sub_module = DataPlaneModule(plane=self.plane, name=name, folder_path=path, parent=self)
                        rp.extend(sub_module.get_resource_providers())
                    else:
                        readme_path = _search_readme_md_path(path, search_parent=True)
                        rp.append(ResourceProvider(name, path, readme_path, swagger_module=self))
        return rp


def _search_readme_md_path(path, search_parent=False):
    if search_parent:
        readme_path = os.path.join(os.path.dirname(path), 'readme.md')
        if os.path.exists(readme_path):
            return readme_path

    readme_path = os.path.join(path, 'readme.md')
    if os.path.exists(readme_path):
        return readme_path

    # find in sub directory
    for name in os.listdir(path):
        sub_path = os.path.join(path, name)
        if os.path.isdir(sub_path):
            readme_path = _search_readme_md_path(sub_path)
            if readme_path is not None:
                return readme_path
    return None




_templates = None


def get_templates():
    global _templates
    if _templates is None:
        import os
        from jinja2 import Environment, FileSystemLoader
        from ._filters import custom_filters
        env = Environment(loader=FileSystemLoader(searchpath=os.path.dirname(os.path.abspath(__file__))))
        env.filters.update(custom_filters)
        _templates = {
            'aaz': {
                '__init__.py': env.get_template("aaz/__init__.py.j2"),
                'profile': {
                    '__init__.py': env.get_template("aaz/profile/__init__.py.j2"),
                },
                'group': {
                    '__cmd_group.py': env.get_template("aaz/group/__cmd_group.py.j2"),
                    '__init__.py': env.get_template("aaz/group/__init__.py.j2"),
                },
                'command': {
                    '_cmd.py': env.get_template("aaz/command/_cmd.py.j2"),
                }
            },
            'extension': {
                "HISTORY.rst": env.get_template("extension/HISTORY.rst.j2"),
                "README.md": env.get_template("extension/README.md.j2"),
                "setup.cfg": env.get_template("extension/setup.cfg.j2"),
                "setup.py": env.get_template("extension/setup.py.j2"),
                "azext_": {
                    "__init__.py": env.get_template("extension/azext_/__init__.py.j2"),
                    "_help.py": env.get_template("extension/azext_/_help.py.j2"),
                    "_params.py": env.get_template("extension/azext_/_params.py.j2"),
                    # "azext_metadata.json": env.get_template("extension/azext_/azext_metadata.json.j2"),
                    "commands.py": env.get_template("extension/azext_/commands.py.j2"),
                    "custom.py": env.get_template("extension/azext_/custom.py.j2"),
                    "tests": {
                        "__init__.py": env.get_template("extension/azext_/tests/__init__.py.j2"),
                        "profile": {
                            "__init__.py": env.get_template("extension/azext_/tests/profile/__init__.py.j2"),
                            "test_.py": env.get_template("extension/azext_/tests/profile/test_.py.j2"),
                        }
                    }
                }
            },
            'main': {
                "__init__.py": env.get_template("main/__init__.py.j2"),
                "_help.py": env.get_template("main/_help.py.j2"),
                "_params.py": env.get_template("main/_params.py.j2"),
                "commands.py": env.get_template("main/commands.py.j2"),
                "custom.py": env.get_template("main/custom.py.j2"),
                "tests": {
                    "__init__.py": env.get_template("main/tests/__init__.py.j2"),
                    "profile": {
                        "__init__.py": env.get_template("main/tests/profile/__init__.py.j2"),
                        "test_.py": env.get_template("main/tests/profile/test_.py.j2"),
                    }
                }
            },
        }
    return _templates



from jinja2.filters import pass_environment
from utils.stage import AAZStageEnum
from utils.case import to_camel_case, to_snack_case
import re


@pass_environment
def camel_case(env, name):
    return to_camel_case(name)


@pass_environment
def snake_case(env, name):
    return to_snack_case(name)


@pass_environment
def is_experimental(env, stage):
    return stage == AAZStageEnum.Experimental


@pass_environment
def is_preview(env, stage):
    return stage == AAZStageEnum.Preview


@pass_environment
def is_stable(env, stage):
    return stage == AAZStageEnum.Stable


@pass_environment
def constant_convert(env, data):
    if isinstance(data, str):
        return f'"%s"' % data.replace('"', '\\"').replace('\n', '\\n')
    elif isinstance(data, (int, float, bool)):
        return f"{data}"
    elif data is None:
        return f"None"
    elif isinstance(data, list):
        return '[%s]' % ', '.join(constant_convert(env, x) for x in data)
    elif isinstance(data, dict):
        return '{%s}' % ', '.join(
            ': '.join([constant_convert(env, k), constant_convert(env, v)])
            for k, v in data.items()
        )
    else:
        raise NotImplementedError()


_PYTHON_BUILD_IN_KEYWORDS = (
    'False', 'await', 'else', 'import', 'pass', 'None', 'break', 'except', 'in', 'raise', 'True', 'class', 'finally',
    'is', 'return', 'and', 'continue', 'for', 'lambda', 'try', 'as', 'def', 'from', 'nonlocal', 'while', 'assert',
    'del', 'global', 'not', 'with', 'async', 'elif', 'if', 'or', 'yield'
)


@pass_environment
def get_prop(env, data):
    assert isinstance(data, str)
    if re.match('^[0-9].*$', data) or data in _PYTHON_BUILD_IN_KEYWORDS:
        return f'[{constant_convert(env, data)}]'
    else:
        return f'.{data}'


custom_filters = {
    "camel_case": camel_case,
    "snake_case": snake_case,
    "is_experimental": is_experimental,
    "is_preview": is_preview,
    "is_stable": is_stable,
    "constant_convert": constant_convert,
    "get_prop": get_prop
}



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from ._module import CLIModule
from ._profile import CLIViewProfile
from ._command_group import CLIViewCommandGroup
from ._command import CLIViewCommand
